"""OT3 Hardware Controller Backend."""

from __future__ import annotations
import asyncio
from contextlib import asynccontextmanager
import logging
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Sequence,
    AsyncIterator,
    cast,
    Set,
    Union,
    Mapping,
)

from opentrons.config.types import OT3Config, GantryLoad
from opentrons.config import gripper_config

from opentrons.hardware_control.module_control import AttachedModulesControl
from opentrons.hardware_control import modules
from opentrons.hardware_control.types import (
    BoardRevision,
    Axis,
    HepaFanState,
    HepaUVState,
    OT3Mount,
    OT3AxisMap,
    CurrentConfig,
    InstrumentProbeType,
    MotorStatus,
    UpdateStatus,
    UpdateState,
    SubSystem,
    SubSystemState,
    TipStateType,
    GripperJawState,
    HardwareFeatureFlags,
    StatusBarState,
    EstopOverallStatus,
    EstopState,
    EstopPhysicalStatus,
    HardwareEventHandler,
    HardwareEventUnsubscriber,
)

from opentrons_shared_data.pipette.types import PipetteName, PipetteModel
from opentrons_shared_data.pipette import (
    pipette_load_name_conversions as pipette_load_name,
    load_data as load_pipette_data,
)
from opentrons_shared_data.gripper.gripper_definition import GripperModel
from opentrons.hardware_control.dev_types import (
    PipetteSpec,
    GripperSpec,
    AttachedPipette,
    AttachedGripper,
    OT3AttachedInstruments,
)
from opentrons.util.async_helpers import ensure_yield
from .types import HWStopCondition
from .flex_protocol import FlexBackend
from opentrons_hardware.firmware_bindings.constants import SensorId
from opentrons_hardware.sensors.types import SensorDataType

log = logging.getLogger(__name__)

AXIS_TO_SUBSYSTEM = {
    Axis.X: SubSystem.gantry_x,
    Axis.Y: SubSystem.gantry_y,
    Axis.Z_L: SubSystem.head,
    Axis.Z_R: SubSystem.head,
    Axis.Z_G: SubSystem.gripper,
    Axis.G: SubSystem.gripper,
    Axis.P_L: SubSystem.pipette_left,
    Axis.P_R: SubSystem.pipette_right,
}


def coalesce_move_segments(
    origin: Dict[Axis, float], targets: List[Dict[Axis, float]]
) -> Dict[Axis, float]:
    for target in targets:
        for axis, increment in target.items():
            origin[axis] += increment
    return origin


def axis_pad(positions: Dict[Axis, float], default_value: float) -> Dict[Axis, float]:
    return {ax: positions.get(ax, default_value) for ax in Axis.node_axes()}


class OT3Simulator(FlexBackend):
    """OT3 Hardware Controller Backend."""

    _position: Dict[Axis, float]
    _encoder_position: Dict[Axis, float]
    _motor_status: Dict[Axis, MotorStatus]
    _engaged_axes: Dict[Axis, bool]

    @classmethod
    async def build(
        cls,
        attached_instruments: Dict[OT3Mount, Dict[str, Optional[str]]],
        attached_modules: Dict[str, List[modules.SimulatingModule]],
        config: OT3Config,
        loop: asyncio.AbstractEventLoop,
        strict_attached_instruments: bool = True,
        feature_flags: Optional[HardwareFeatureFlags] = None,
    ) -> OT3Simulator:
        """Create the OT3Simulator instance.

        Args:
            config: Robot configuration

        Returns:
            Instance.
        """
        return cls(
            attached_instruments,
            attached_modules,
            config,
            loop,
            strict_attached_instruments,
            feature_flags,
        )

    def __init__(
        self,
        attached_instruments: Dict[OT3Mount, Dict[str, Optional[str]]],
        attached_modules: Dict[str, List[modules.SimulatingModule]],
        config: OT3Config,
        loop: asyncio.AbstractEventLoop,
        strict_attached_instruments: bool = True,
        feature_flags: Optional[HardwareFeatureFlags] = None,
    ) -> None:
        """Construct.

        Args:
            config: Robot configuration
            driver: The Can Driver
        """
        self._configuration = config
        self._loop = loop
        self._strict_attached = bool(strict_attached_instruments)
        self._stubbed_attached_modules = attached_modules
        self._update_required = False
        self._initialized = False
        self._lights = {"button": False, "rails": False}
        self._gear_motor_position: Dict[Axis, float] = {}
        self._engaged_axes: Dict[Axis, bool] = {}
        self._feature_flags = feature_flags or HardwareFeatureFlags()

        def _sanitize_attached_instrument(
            mount: OT3Mount, passed_ai: Optional[Dict[str, Optional[str]]] = None
        ) -> Union[PipetteSpec, GripperSpec]:
            if mount is OT3Mount.GRIPPER:
                gripper_spec: GripperSpec = {"model": None, "id": None}
                if passed_ai and passed_ai.get("model"):
                    gripper_spec["model"] = GripperModel.v1
                    gripper_spec["id"] = passed_ai.get("id")
                return gripper_spec

            # TODO (lc 12-5-2022) need to not always pass in defaults here
            # but doing it to satisfy linter errors for now.
            pipette_spec: PipetteSpec = {"model": None, "id": None}
            if not passed_ai or not passed_ai.get("model"):
                return pipette_spec

            if pipette_load_name.supported_pipette(
                cast(PipetteModel, passed_ai["model"])
            ):
                pipette_spec["model"] = cast(PipetteModel, passed_ai.get("model"))
                pipette_spec["id"] = passed_ai.get("id")
                return pipette_spec
            # TODO (lc 12-05-2022) When the time comes we should properly
            # support backwards compatibility
            raise KeyError(
                "If you specify attached_instruments, the model "
                "should be pipette names or pipette models, but "
                f'{passed_ai["model"]} is not'
            )

        self._attached_instruments = {
            m: _sanitize_attached_instrument(m, attached_instruments.get(m))
            for m in OT3Mount
        }
        self._module_controls: Optional[AttachedModulesControl] = None
        self._position = self._get_home_position()
        self._encoder_position = self._get_home_position()
        self._motor_status = {}
        axes = set((Axis.Z_L, Axis.Z_R, Axis.X, Axis.Y))
        if self._attached_instruments[OT3Mount.LEFT].get("model", None):
            axes.add(Axis.P_L)
        if self._attached_instruments[OT3Mount.RIGHT].get("model", None):
            axes.add(Axis.P_L)
        if self._attached_instruments.get(
            OT3Mount.GRIPPER
        ) and self._attached_instruments[OT3Mount.GRIPPER].get("model", None):
            axes.update((Axis.G, Axis.Z_G))
        self._present_axes = axes
        self._current_settings: Optional[OT3AxisMap[CurrentConfig]] = None
        self._sim_jaw_state = GripperJawState.HOMED_READY
        self._sim_tip_state: Dict[OT3Mount, Optional[bool]] = {
            mount: False if self._attached_instruments[mount] else None
            for mount in [OT3Mount.LEFT, OT3Mount.RIGHT]
        }
        self._sim_gantry_load = GantryLoad.LOW_THROUGHPUT
        self._sim_status_bar_state = StatusBarState.IDLE
        self._sim_estop_state = EstopState.DISENGAGED
        self._sim_estop_left_state = EstopPhysicalStatus.DISENGAGED
        self._sim_estop_right_state = EstopPhysicalStatus.DISENGAGED

    async def get_serial_number(self) -> Optional[str]:
        return "simulator"

    @asynccontextmanager
    async def restore_system_constraints(self) -> AsyncIterator[None]:
        log.debug("Simulating saving system constraints")
        try:
            yield
        finally:
            log.debug("Simulating restoring system constraints")

    def update_constraints_for_gantry_load(self, gantry_load: GantryLoad) -> None:
        self._sim_gantry_load = gantry_load

    def update_constraints_for_calibration_with_gantry_load(
        self,
        gantry_load: GantryLoad,
    ) -> None:
        self._sim_gantry_load = gantry_load

    def update_constraints_for_plunger_acceleration(
        self, mount: OT3Mount, acceleration: float, gantry_load: GantryLoad
    ) -> None:
        self._sim_gantry_load = gantry_load

    @property
    def initialized(self) -> bool:
        """True when the hardware controller has initialized and is ready."""
        return self._initialized

    @initialized.setter
    def initialized(self, value: bool) -> None:
        self._initialized = value

    @property
    def gear_motor_position(self) -> Optional[float]:
        return self._gear_motor_position.get(Axis.Q, None)

    @property
    def board_revision(self) -> BoardRevision:
        """Get the board revision"""
        return BoardRevision.UNKNOWN

    @property
    def module_controls(self) -> AttachedModulesControl:
        """Get the module controls."""
        if self._module_controls is None:
            raise AttributeError("Module controls not found.")
        return self._module_controls

    @module_controls.setter
    def module_controls(self, module_controls: AttachedModulesControl) -> None:
        """Set the module controls"""
        self._module_controls = module_controls

    @ensure_yield
    async def update_to_default_current_settings(self, gantry_load: GantryLoad) -> None:
        self._gantry_load = gantry_load

    def update_feature_flags(self, feature_flags: HardwareFeatureFlags) -> None:
        """Update the hardware feature flags used by the hardware controller."""
        self._feature_flags = feature_flags

    def _handle_motor_status_update(self, response: Dict[Axis, float]) -> None:
        self._position.update(response)
        self._encoder_position.update(response)
        self._motor_status.update(
            (node, MotorStatus(True, True)) for node in response.keys()
        )

    @ensure_yield
    async def update_motor_status(self) -> None:
        """Retreieve motor and encoder status and position from all present devices"""
        if not self._motor_status:
            # Simulate condition at boot, status would not be ok
            self._motor_status.update(
                (axis, MotorStatus(False, False)) for axis in self._present_axes
            )
        else:
            self._motor_status.update(
                (axis, MotorStatus(True, True)) for axis in self._present_axes
            )

    @ensure_yield
    async def update_motor_estimation(self, axes: Sequence[Axis]) -> None:
        """Update motor position estimation for commanded axes, and update cache of data."""
        # Simulate conditions as if there are no stalls, aka do nothing
        return None

    def _get_motor_status(
        self, axes: Sequence[Axis]
    ) -> Dict[Axis, Optional[MotorStatus]]:
        return {ax: self._motor_status.get(ax) for ax in axes}

    def get_invalid_motor_axes(self, axes: Sequence[Axis]) -> List[Axis]:
        """Get axes that currently do not have the motor-ok flag."""
        return [
            ax
            for ax, status in self._get_motor_status(axes).items()
            if not status or not status.motor_ok
        ]

    def get_invalid_encoder_axes(self, axes: Sequence[Axis]) -> List[Axis]:
        """Get axes that currently do not have the encoder-ok flag."""
        return [
            ax
            for ax, status in self._get_motor_status(axes).items()
            if not status or not status.encoder_ok
        ]

    def check_motor_status(self, axes: Sequence[Axis]) -> bool:
        return len(self.get_invalid_motor_axes(axes)) == 0

    def check_encoder_status(self, axes: Sequence[Axis]) -> bool:
        return len(self.get_invalid_encoder_axes(axes)) == 0

    async def update_position(self) -> OT3AxisMap[float]:
        """Get the current position."""
        return axis_pad(self._position, 0.0)

    async def update_encoder_position(self) -> OT3AxisMap[float]:
        """Get the encoder current position."""
        return axis_pad(self._encoder_position, 0.0)

    @ensure_yield
    async def liquid_probe(
        self,
        mount: OT3Mount,
        max_p_distance: float,
        mount_speed: float,
        plunger_speed: float,
        threshold_pascals: float,
        plunger_impulse_time: float,
        num_baseline_reads: int,
        probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
        force_both_sensors: bool = False,
        response_queue: Optional[
            asyncio.Queue[Dict[SensorId, List[SensorDataType]]]
        ] = None,
    ) -> float:
        z_axis = Axis.by_mount(mount)
        pos = self._position
        self._position.update(pos)
        self._encoder_position.update(pos)
        return self._position[z_axis]

    @ensure_yield
    async def move(
        self,
        origin: Dict[Axis, float],
        target: Dict[Axis, float],
        speed: Optional[float] = None,
        stop_condition: HWStopCondition = HWStopCondition.none,
        nodes_in_moves_only: bool = True,
    ) -> None:
        """Move to a position.

        Args:
            target_position: Map of axis to position.
            home_flagged_axes: Whether to home afterwords.
            speed: Optional speed
            axis_max_speeds: Optional map of axis to speed.

        Returns:
            None
        """
        for ax in origin:
            self._engaged_axes[ax] = True
        self._position.update(target)
        self._encoder_position.update(target)

    @ensure_yield
    async def home(
        self, axes: Sequence[Axis], gantry_load: GantryLoad
    ) -> OT3AxisMap[float]:
        """Home axes.

        Args:
            axes: Optional list of axes.

        Returns:
            Homed position.
        """
        if axes:
            homed = axes
        else:
            homed = list(iter(self._position.keys()))
        for h in homed:
            self._position[h] = self._get_home_position()[h]
            self._motor_status[h] = MotorStatus(True, True)
            self._engaged_axes[h] = True
        return axis_pad(self._position, 0.0)

    @ensure_yield
    async def gripper_grip_jaw(
        self,
        duty_cycle: float,
        expected_displacement: float,
        stop_condition: HWStopCondition = HWStopCondition.none,
        stay_engaged: bool = True,
    ) -> None:
        """Move gripper inward."""
        self._sim_jaw_state = GripperJawState.GRIPPING
        self._encoder_position[Axis.G] = expected_displacement

    @ensure_yield
    async def gripper_home_jaw(self, duty_cycle: float) -> None:
        """Move gripper outward."""
        self._motor_status[Axis.G] = MotorStatus(True, True)
        self._encoder_position[Axis.G] = self._get_home_position()[Axis.G]
        self._sim_jaw_state = GripperJawState.HOMED_READY

    @ensure_yield
    async def gripper_hold_jaw(
        self,
        encoder_position_um: int,
    ) -> None:
        self._encoder_position[Axis.G] = encoder_position_um / 1000.0
        self._sim_jaw_state = GripperJawState.HOLDING

    async def get_jaw_state(self) -> GripperJawState:
        """Get the state of the gripper jaw."""
        return self._sim_jaw_state

    async def tip_action(
        self, origin: Dict[Axis, float], targets: List[Tuple[Dict[Axis, float], float]]
    ) -> None:
        self._gear_motor_position.update(
            coalesce_move_segments(origin, [target[0] for target in targets])
        )
        await asyncio.sleep(0)

    async def home_tip_motors(
        self,
        distance: float,
        velocity: float,
        back_off: bool = True,
    ) -> None:
        pass

    def _attached_to_mount(
        self, mount: OT3Mount, expected_instr: Optional[PipetteName]
    ) -> OT3AttachedInstruments:
        init_instr = self._attached_instruments.get(mount, {"model": None, "id": None})
        if mount is OT3Mount.GRIPPER:
            return self._attached_gripper_to_mount(cast(GripperSpec, init_instr))
        return self._attached_pipette_to_mount(
            mount, cast(PipetteSpec, init_instr), expected_instr
        )

    def _attached_gripper_to_mount(self, init_instr: GripperSpec) -> AttachedGripper:
        found_model = init_instr["model"]
        if found_model:
            return {
                "config": gripper_config.load(GripperModel.v1),
                "id": init_instr["id"],
            }
        else:
            return {"config": None, "id": None}

    def _attached_pipette_to_mount(
        self,
        mount: OT3Mount,
        init_instr: PipetteSpec,
        expected_instr: Optional[PipetteName],
    ) -> AttachedPipette:
        found_model = init_instr["model"]

        # TODO (lc 12-05-2022) When the time comes, we should think about supporting
        # backwards compatability -- hopefully not relying on config keys only,
        # but TBD.
        if expected_instr and not pipette_load_name.supported_pipette(
            cast(PipetteModel, expected_instr)
        ):
            raise RuntimeError(
                f"mount {mount.name} requested a {expected_instr} which is not supported on the OT3"
            )
        if found_model and expected_instr and (expected_instr not in found_model):
            if self._strict_attached:
                raise RuntimeError(
                    "mount {}: expected instrument {} but got {}".format(
                        mount.name, expected_instr, found_model
                    )
                )
            else:
                converted_name = pipette_load_name.convert_pipette_name(expected_instr)
                return {
                    "config": load_pipette_data.load_definition(
                        converted_name.pipette_type,
                        converted_name.pipette_channels,
                        converted_name.pipette_version,
                    ),
                    "id": None,
                }
        if found_model and init_instr["id"] is not None:
            # Instrument detected matches instrument expected (note:
            # "instrument detected" means passed as an argument to the
            # constructor of this class)

            # OR Instrument detected and no expected instrument specified

            found_model_version = ""
            if found_model.find("flex") > -1:
                found_model = found_model.replace("_flex", "")  # type: ignore
                found_model_version = f"{init_instr['id'][4]}.{init_instr['id'][5]}"
            converted_name = pipette_load_name.convert_pipette_model(
                found_model, found_model_version
            )
            return {
                "config": load_pipette_data.load_definition(
                    converted_name.pipette_type,
                    converted_name.pipette_channels,
                    converted_name.pipette_version,
                ),
                "id": init_instr["id"],
            }
        elif expected_instr:
            # Expected instrument specified and no instrument detected
            converted_name = pipette_load_name.convert_pipette_name(expected_instr)
            return {
                "config": load_pipette_data.load_definition(
                    converted_name.pipette_type,
                    converted_name.pipette_channels,
                    converted_name.pipette_version,
                ),
                "id": None,
            }
        else:
            # No instrument detected or expected
            return {"config": None, "id": None}

    @ensure_yield
    async def get_attached_instruments(
        self, expected: Mapping[OT3Mount, Optional[PipetteName]]
    ) -> Mapping[OT3Mount, OT3AttachedInstruments]:
        """Get attached instruments.

        Args:
            expected: Which mounts are expected.

        Returns:
            A map of mount to pipette name.
        """
        return {
            mount: self._attached_to_mount(mount, expected.get(mount))
            for mount in OT3Mount
        }

    @ensure_yield
    async def get_limit_switches(self) -> OT3AxisMap[bool]:
        """Get the state of the gantry's limit switches on each axis."""
        return {}

    @ensure_yield
    async def set_active_current(self, axis_currents: OT3AxisMap[float]) -> None:
        """Set the active current.

        Args:
            axis_currents: Axes' currents

        Returns:
            None
        """
        return None

    @asynccontextmanager
    async def motor_current(
        self,
        run_currents: Optional[OT3AxisMap[float]] = None,
        hold_currents: Optional[OT3AxisMap[float]] = None,
    ) -> AsyncIterator[None]:
        """Save the current."""
        yield

    @asynccontextmanager
    async def restore_z_r_run_current(self) -> AsyncIterator[None]:
        """
        Temporarily restore the active current ONLY when homing or
        retracting the Z_R axis while the 96-channel is attached.
        """
        yield

    @asynccontextmanager
    async def increase_z_l_hold_current(self) -> AsyncIterator[None]:
        """
        Temporarily increase the hold current when engaging the Z_L axis
        while the 96-channel is attached
        """
        yield

    @ensure_yield
    async def watch(self, loop: asyncio.AbstractEventLoop) -> None:
        new_mods_at_ports = []
        for mod_name, list_of_modules in self._stubbed_attached_modules.items():
            for module_details in list_of_modules:
                new_mods_at_ports.append(
                    modules.SimulatingModuleAtPort(
                        port=f"/dev/ot_module_sim_{mod_name}{str(module_details.serial_number)}",
                        name=mod_name,
                        serial_number=module_details.serial_number,
                        model=module_details.model,
                    )
                )
        await self.module_controls.register_modules(new_mods_at_ports=new_mods_at_ports)

    @property
    def axis_bounds(self) -> OT3AxisMap[Tuple[float, float]]:
        """Get the axis bounds."""
        # TODO (AL, 2021-11-18): The bounds need to be defined
        phony_bounds = (0, 10000)
        return {
            Axis.Z_R: phony_bounds,
            Axis.Z_L: phony_bounds,
            Axis.P_L: phony_bounds,
            Axis.P_R: phony_bounds,
            Axis.Y: phony_bounds,
            Axis.X: phony_bounds,
            Axis.Z_G: phony_bounds,
            Axis.Q: phony_bounds,
        }

    @property
    def fw_version(self) -> Dict[SubSystem, int]:
        """Get the firmware version."""
        return {AXIS_TO_SUBSYSTEM[axis]: 0 for axis in self._present_axes}

    def axis_is_present(self, axis: Axis) -> bool:
        return axis in self._present_axes

    @property
    def update_required(self) -> bool:
        return self._update_required

    @update_required.setter
    def update_required(self, value: bool) -> None:
        if value != self._update_required:
            log.info(f"Firmware Update Flag set {self._update_required} -> {value}")
            self._update_required = value

    async def update_firmware(
        self,
        subsystems: Set[SubSystem],
        force: bool = False,
    ) -> AsyncIterator[UpdateStatus]:
        """Updates the firmware on the OT3."""
        for subsystem in subsystems:
            yield UpdateStatus(
                subsystem=subsystem, state=UpdateState.done, progress=100
            )

    def engaged_axes(self) -> OT3AxisMap[bool]:
        """Get engaged axes."""
        return self._engaged_axes

    async def update_engaged_axes(self) -> None:
        """Update engaged axes."""
        return None

    async def is_motor_engaged(self, axis: Axis) -> bool:
        if axis not in self._engaged_axes.keys():
            return False
        return self._engaged_axes[axis]

    @ensure_yield
    async def disengage_axes(self, axes: List[Axis]) -> None:
        """Disengage axes."""
        for ax in axes:
            self._engaged_axes.update({ax: False})
        return None

    @ensure_yield
    async def engage_axes(self, axes: List[Axis]) -> None:
        """Engage axes."""
        for ax in axes:
            self._engaged_axes.update({ax: True})
        return None

    @ensure_yield
    async def set_lights(self, button: Optional[bool], rails: Optional[bool]) -> None:
        """Set the light states."""
        # Simulate how the real driver does this - there's no button so it's always false
        if rails is not None:
            self._lights["rails"] = rails

    @ensure_yield
    async def get_lights(self) -> Dict[str, bool]:
        """Get the light state."""
        return self._lights

    def pause(self) -> None:
        """Pause the controller activity."""
        return None

    def resume(self) -> None:
        """Resume the controller activity."""
        return None

    @ensure_yield
    async def halt(self) -> None:
        """Halt the motors."""
        return None

    @ensure_yield
    async def probe(self, axis: Axis, distance: float) -> OT3AxisMap[float]:
        """Probe."""
        return {}

    @ensure_yield
    async def clean_up(self) -> None:
        """Clean up."""
        pass

    @staticmethod
    def _get_home_position() -> Dict[Axis, float]:
        return {
            Axis.Z_L: 0,
            Axis.Z_R: 0,
            Axis.X: 0,
            Axis.Y: 0,
            Axis.P_L: 0,
            Axis.P_R: 0,
            Axis.Z_G: 0,
            Axis.G: 0,
        }

    @staticmethod
    def home_position() -> OT3AxisMap[float]:
        return OT3Simulator._get_home_position()

    @ensure_yield
    async def capacitive_probe(
        self,
        mount: OT3Mount,
        moving: Axis,
        distance_mm: float,
        speed_mm_per_s: float,
        sensor_threshold_pf: float,
        probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
    ) -> bool:
        self._position[moving] += distance_mm
        return True

    @ensure_yield
    async def capacitive_pass(
        self,
        mount: OT3Mount,
        moving: Axis,
        distance_mm: float,
        speed_mm_per_s: float,
        probe: InstrumentProbeType,
    ) -> List[float]:
        self._position[moving] += distance_mm
        return []

    @property
    def subsystems(self) -> Dict[SubSystem, SubSystemState]:
        return {
            AXIS_TO_SUBSYSTEM[axis]: SubSystemState(
                ok=True,
                current_fw_version=1,
                next_fw_version=1,
                fw_update_needed=False,
                current_fw_sha="simulated",
                pcba_revision="A1",
                update_state=None,
            )
            for axis in self._present_axes
        }

    async def get_tip_status(
        self,
        mount: OT3Mount,
        follow_singular_sensor: Optional[InstrumentProbeType] = None,
    ) -> TipStateType:
        return TipStateType(self._sim_tip_state[mount])

    def current_tip_state(self, mount: OT3Mount) -> Optional[bool]:
        return self._sim_tip_state[mount]

    async def update_tip_detector(self, mount: OT3Mount, sensor_count: int) -> None:
        pass

    async def teardown_tip_detector(self, mount: OT3Mount) -> None:
        pass

    async def set_status_bar_state(self, state: StatusBarState) -> None:
        self._sim_status_bar_state = state
        await asyncio.sleep(0)

    async def set_status_bar_enabled(self, enabled: bool) -> None:
        await asyncio.sleep(0)

    def get_status_bar_state(self) -> StatusBarState:
        return self._sim_status_bar_state

    @property
    def estop_status(self) -> EstopOverallStatus:
        return EstopOverallStatus(
            state=self._sim_estop_state,
            left_physical_state=self._sim_estop_left_state,
            right_physical_state=self._sim_estop_right_state,
        )

    def estop_acknowledge_and_clear(self) -> EstopOverallStatus:
        """Attempt to acknowledge an Estop event and clear the status.

        Returns the estop status after clearing the status."""
        self._sim_estop_state = EstopState.DISENGAGED
        self._sim_estop_left_state = EstopPhysicalStatus.DISENGAGED
        self._sim_estop_right_state = EstopPhysicalStatus.DISENGAGED
        return self.estop_status

    def get_estop_state(self) -> EstopState:
        return self._sim_estop_state

    def add_estop_callback(self, cb: HardwareEventHandler) -> HardwareEventUnsubscriber:
        return lambda: None

    def check_gripper_position_within_bounds(
        self,
        expected_grip_width: float,
        grip_width_uncertainty_wider: float,
        grip_width_uncertainty_narrower: float,
        jaw_width: float,
        max_allowed_grip_error: float,
        hard_limit_lower: float,
        hard_limit_upper: float,
    ) -> None:
        # This is a (pretty bad) simulation of the gripper actually gripping something,
        # but it should work.
        self._encoder_position[Axis.G] = (hard_limit_upper - jaw_width) / 2

    async def set_hepa_fan_state(self, fan_on: bool, duty_cycle: int) -> bool:
        return False

    async def get_hepa_fan_state(self) -> Optional[HepaFanState]:
        return None

    async def set_hepa_uv_state(self, light_on: bool, timeout_s: int) -> bool:
        return False

    async def get_hepa_uv_state(self) -> Optional[HepaUVState]:
        return None

    def _update_tip_state(self, mount: OT3Mount, status: bool) -> None:
        """This is something we only use in the simulator.
        It is required so that PE simulations using ot3api don't break."""
        self._sim_tip_state[mount] = status



""" Entrypoint for the OT2 hardware control server

This server listens on a variety of different interfaces and runs a hardware
controller. It is the only thing that is allowed to talk to the OT2's hardware,
including the smoothie, the camera, and the various GPIO interfaces.

This server can only be run on a system with a connected smoothie board, and
_should_ only be run on an OT-2.
"""

import argparse
import asyncio
import logging
from typing import Optional, Dict, Any

from . import API
from .types import HardwareFeatureFlags
from opentrons.config import robot_configs as rc
from opentrons.config.types import RobotConfig

LOG = logging.getLogger("opentrons.hardware_control.__main__")


def exception_handler(loop: asyncio.AbstractEventLoop, context: Dict[str, Any]) -> None:
    message = ""
    if "exception" in context:
        message += f'exception: {repr(context["exception"])}'
    if "future" in context:
        message += f' while running future {repr(context["future"])}'
    if "protocol" in context:
        message += f' from protocol {repr(context["protocol"])}'
    if "transport" in context:
        message += f' from transport {repr(context["transport"])}'
    if "socket" in context:
        message += f' from socket {repr(context["socket"])}'
    LOG.error(f"Unhandled exception in event loop: {message}")
    loop.default_exception_handler(context)


async def arun(
    config: Optional[RobotConfig] = None, port: Optional[str] = None
) -> None:
    """Asynchronous entrypoint for the server

    :param config: Optional config override
    :param port: Optional smoothie port override
    """
    rconf = config or rc.load()
    hc = await API.build_hardware_controller(  # noqa: F841
        config=rconf, port=port, feature_flags=HardwareFeatureFlags.build_from_ff()
    )


def run(config: Optional[RobotConfig] = None, port: Optional[str] = None) -> None:
    """Synchronous entrypoint for the server.

    Mostly builds a loop and calls arun.

    :param config: Optional config override
    :param port: Optional Smoothie port override
    """
    loop = asyncio.get_event_loop()
    loop.set_exception_handler(exception_handler)
    loop.run_until_complete(arun(config, port))
    loop.run_forever()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Opentrons hardware control server")
    parser.add_argument(
        "-s",
        "--smoothie-port",
        help="Port on which to talk to smoothie, autodetected by default",
        default=None,
    )
    args = parser.parse_args()
    run(port=args.port)



"""Drop tip in place command request, result, and implementation models."""
from __future__ import annotations
from pydantic import Field, BaseModel
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from .pipetting_common import PipetteIdMixin, TipPhysicallyAttachedError
from ..errors.exceptions import TipAttachedError
from ..errors.error_occurrence import ErrorOccurrence
from ..resources.model_utils import ModelUtils
from ..state import update_types

if TYPE_CHECKING:
    from ..execution import TipHandler, GantryMover


DropTipInPlaceCommandType = Literal["dropTipInPlace"]


class DropTipInPlaceParams(PipetteIdMixin):
    """Payload required to drop a tip in place."""

    homeAfter: Optional[bool] = Field(
        None,
        description=(
            "Whether to home this pipette's plunger after dropping the tip."
            " You should normally leave this unspecified to let the robot choose"
            " a safe default depending on its hardware."
        ),
    )


class DropTipInPlaceResult(BaseModel):
    """Result data from the execution of a DropTipInPlace command."""

    pass


_ExecuteReturn = (
    SuccessData[DropTipInPlaceResult] | DefinedErrorData[TipPhysicallyAttachedError]
)


class DropTipInPlaceImplementation(
    AbstractCommandImpl[DropTipInPlaceParams, _ExecuteReturn]
):
    """Drop tip in place command implementation."""

    def __init__(
        self,
        tip_handler: TipHandler,
        model_utils: ModelUtils,
        gantry_mover: GantryMover,
        **kwargs: object,
    ) -> None:
        self._tip_handler = tip_handler
        self._model_utils = model_utils
        self._gantry_mover = gantry_mover

    async def execute(self, params: DropTipInPlaceParams) -> _ExecuteReturn:
        """Drop a tip using the requested pipette."""
        state_update = update_types.StateUpdate()

        retry_location = await self._gantry_mover.get_position(params.pipetteId)

        try:
            await self._tip_handler.drop_tip(
                pipette_id=params.pipetteId, home_after=params.homeAfter
            )
        except TipAttachedError as exception:
            state_update_if_false_positive = update_types.StateUpdate()
            state_update_if_false_positive.update_pipette_tip_state(
                pipette_id=params.pipetteId, tip_geometry=None
            )
            error = TipPhysicallyAttachedError(
                id=self._model_utils.generate_id(),
                createdAt=self._model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=self._model_utils.generate_id(),
                        createdAt=self._model_utils.get_timestamp(),
                        error=exception,
                    )
                ],
                errorInfo={"retryLocation": retry_location},
            )
            return DefinedErrorData(
                public=error,
                state_update=state_update,
                state_update_if_false_positive=state_update_if_false_positive,
            )
        else:
            state_update.update_pipette_tip_state(
                pipette_id=params.pipetteId, tip_geometry=None
            )
            return SuccessData(public=DropTipInPlaceResult(), state_update=state_update)


class DropTipInPlace(
    BaseCommand[DropTipInPlaceParams, DropTipInPlaceResult, TipPhysicallyAttachedError]
):
    """Drop tip in place command model."""

    commandType: DropTipInPlaceCommandType = "dropTipInPlace"
    params: DropTipInPlaceParams
    result: Optional[DropTipInPlaceResult]

    _ImplementationCls: Type[
        DropTipInPlaceImplementation
    ] = DropTipInPlaceImplementation


class DropTipInPlaceCreate(BaseCommandCreate[DropTipInPlaceParams]):
    """Drop tip in place command creation request model."""

    commandType: DropTipInPlaceCommandType = "dropTipInPlace"
    params: DropTipInPlaceParams

    _CommandCls: Type[DropTipInPlace] = DropTipInPlace



"""Load pipette command request, result, and implementation models."""
from __future__ import annotations

from opentrons.protocol_engine.state.update_types import StateUpdate
from opentrons_shared_data.pipette.pipette_load_name_conversions import (
    convert_to_pipette_name_type,
)
from opentrons_shared_data.pipette.types import PipetteGenerationType
from opentrons_shared_data.robot import user_facing_robot_type
from opentrons_shared_data.robot.types import RobotTypeEnum
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons.types import MountType

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence
from ..errors import InvalidSpecificationForRobotTypeError, InvalidLoadPipetteSpecsError

if TYPE_CHECKING:
    from ..execution import EquipmentHandler
    from ..state.state import StateView


LoadPipetteCommandType = Literal["loadPipette"]


class LoadPipetteParams(BaseModel):
    """Payload needed to load a pipette on to a mount."""

    pipetteName: PipetteNameType = Field(
        ...,
        description="The load name of the pipette to be required.",
    )
    mount: MountType = Field(
        ...,
        description="The mount the pipette should be present on.",
    )
    pipetteId: Optional[str] = Field(
        None,
        description="An optional ID to assign to this pipette. If None, an ID "
        "will be generated.",
    )
    tipOverlapNotAfterVersion: Optional[str] = Field(
        None,
        description="A version of tip overlap data to not exceed. The highest-versioned "
        "tip overlap data that does not exceed this version will be used. Versions are "
        "expressed as vN where N is an integer, counting up from v0. If None, the current "
        "highest version will be used.",
    )
    liquidPresenceDetection: Optional[bool] = Field(
        None,
        description="Enable liquid presence detection for this pipette. Defaults to False.",
    )


class LoadPipetteResult(BaseModel):
    """Result data for executing a LoadPipette."""

    pipetteId: str = Field(
        ...,
        description="An ID to reference this pipette in subsequent commands.",
    )


class LoadPipetteImplementation(
    AbstractCommandImpl[LoadPipetteParams, SuccessData[LoadPipetteResult]]
):
    """Load pipette command implementation."""

    def __init__(
        self, equipment: EquipmentHandler, state_view: StateView, **kwargs: object
    ) -> None:
        self._equipment = equipment
        self._state_view = state_view

    async def execute(
        self, params: LoadPipetteParams
    ) -> SuccessData[LoadPipetteResult]:
        """Check that requested pipette is attached and assign its identifier."""
        pipette_generation = convert_to_pipette_name_type(
            params.pipetteName.value
        ).pipette_generation
        robot_type = RobotTypeEnum.robot_literal_to_enum(
            self._state_view.config.robot_type
        )
        if (
            (
                robot_type == RobotTypeEnum.FLEX
                and pipette_generation != PipetteGenerationType.FLEX
            )
        ) or (
            (
                robot_type == RobotTypeEnum.OT2
                and pipette_generation
                not in [PipetteGenerationType.GEN1, PipetteGenerationType.GEN2]
            )
        ):
            raise InvalidSpecificationForRobotTypeError(
                f"Cannot load a {pipette_generation.value.capitalize()} pipette on "
                f"{user_facing_robot_type(robot_type=self._state_view.config.robot_type, include_article=True)}."
            )

        if params.mount == MountType.EXTENSION:
            raise InvalidLoadPipetteSpecsError(
                "Cannot load a pipette on the EXTENSION mount. Use mount LEFT or RIGHT."
            )

        loaded_pipette = await self._equipment.load_pipette(
            pipette_name=params.pipetteName,
            mount=params.mount,
            pipette_id=params.pipetteId,
            tip_overlap_version=params.tipOverlapNotAfterVersion,
        )

        state_update = StateUpdate()
        state_update.set_load_pipette(
            pipette_id=loaded_pipette.pipette_id,
            pipette_name=params.pipetteName,
            mount=params.mount,
            liquid_presence_detection=params.liquidPresenceDetection,
        )
        state_update.update_pipette_config(
            pipette_id=loaded_pipette.pipette_id,
            serial_number=loaded_pipette.serial_number,
            config=loaded_pipette.static_config,
        )

        return SuccessData(
            public=LoadPipetteResult(pipetteId=loaded_pipette.pipette_id),
            state_update=state_update,
        )


class LoadPipette(BaseCommand[LoadPipetteParams, LoadPipetteResult, ErrorOccurrence]):
    """Load pipette command model."""

    commandType: LoadPipetteCommandType = "loadPipette"
    params: LoadPipetteParams
    result: Optional[LoadPipetteResult]

    _ImplementationCls: Type[LoadPipetteImplementation] = LoadPipetteImplementation


class LoadPipetteCreate(BaseCommandCreate[LoadPipetteParams]):
    """Load pipette command creation request model."""

    commandType: LoadPipetteCommandType = "loadPipette"
    params: LoadPipetteParams

    _CommandCls: Type[LoadPipette] = LoadPipette



"""Simulating AbstractRunner factory."""

from opentrons.hardware_control import API as OT2API, HardwareControlAPI
from opentrons.protocols.api_support import deck_type
from opentrons.protocols.api_support.deck_type import should_load_fixed_trash
from opentrons.protocol_engine import (
    Config as ProtocolEngineConfig,
    DeckType,
    error_recovery_policy,
)
from opentrons.protocol_engine.create_protocol_engine import create_protocol_engine
from opentrons.protocol_reader.protocol_source import ProtocolConfig

from opentrons_shared_data.robot.types import RobotType

from .python_protocol_wrappers import SimulatingContextCreator
from .run_orchestrator import RunOrchestrator
from .protocol_runner import create_protocol_runner, LiveRunner


async def create_simulating_orchestrator(
    robot_type: RobotType, protocol_config: ProtocolConfig
) -> RunOrchestrator:
    """Create a RunOrchestrator wired to a simulating HardwareControlAPI.

    Example:
        ```python
        from pathlib import Path
        from typing import List
        from opentrons.protocol_engine import Command
        from opentrons.protocol_runner import (
            ProtocolType,
            ProtocolFile,
            AbstractRunner,
            create_simulating_orchestrator,
        )

        protocol = ProtocolFile(
            protocol_type=ProtocolType.PYTHON,
            files=[Path("/path/to/protocol.py")],
        )
        orchestrator: RunOrchestrator = await create_simulating_orchestrator()
        commands: List[Command] = await orchestrator.run(protocol)
        ```
    """
    simulating_hardware_api = await _build_hardware_simulator_for_robot_type(
        robot_type=robot_type
    )

    # TODO(mm, 2024-08-06): This home has theoretically been replaced by Protocol Engine
    # `home` commands within the `RunOrchestrator` or `ProtocolRunner`. However, it turns
    # out that this `HardwareControlAPI`-level home is accidentally load-bearing,
    # working around Protocol Engine bugs where *both* layers need to be homed for
    # certain commands to work. https://opentrons.atlassian.net/browse/EXEC-646
    await simulating_hardware_api.home()

    protocol_engine = await create_protocol_engine(
        hardware_api=simulating_hardware_api,
        config=ProtocolEngineConfig(
            robot_type=robot_type,
            deck_type=DeckType(deck_type.for_simulation(robot_type)),
            ignore_pause=True,
            use_virtual_modules=True,
            use_virtual_gripper=True,
            use_simulated_deck_config=True,
            use_virtual_pipettes=True,
        ),
        error_recovery_policy=error_recovery_policy.never_recover,
        load_fixed_trash=should_load_fixed_trash(protocol_config),
    )

    simulating_context_creator = SimulatingContextCreator(
        hardware_api=simulating_hardware_api,
        protocol_engine=protocol_engine,
    )

    runner = create_protocol_runner(
        protocol_config=protocol_config,
        protocol_engine=protocol_engine,
        hardware_api=simulating_hardware_api,
        protocol_context_creator=simulating_context_creator,
    )

    setup_runner = LiveRunner(
        protocol_engine=protocol_engine,
        hardware_api=simulating_hardware_api,
    )

    fixit_runner = LiveRunner(
        protocol_engine=protocol_engine,
        hardware_api=simulating_hardware_api,
    )

    protocol_live_runner = LiveRunner(
        protocol_engine=protocol_engine,
        hardware_api=simulating_hardware_api,
    )

    return RunOrchestrator(
        hardware_api=simulating_hardware_api,
        json_or_python_protocol_runner=runner,
        protocol_engine=protocol_engine,
        setup_runner=setup_runner,
        fixit_runner=fixit_runner,
        protocol_live_runner=protocol_live_runner,
    )


async def _build_hardware_simulator_for_robot_type(
    robot_type: RobotType,
) -> HardwareControlAPI:
    if robot_type == "OT-2 Standard":
        return await OT2API.build_hardware_simulator()
    elif robot_type == "OT-3 Standard":
        # Inline import because OT3API is not present to import on an OT-2 system.
        from opentrons.hardware_control.ot3api import OT3API

        return await OT3API.build_hardware_simulator()



"""File interaction resource provider."""
from datetime import datetime
from typing import List, Optional, Callable, Awaitable, Dict
from pydantic import BaseModel
from ..errors import StorageLimitReachedError


MAXIMUM_CSV_FILE_LIMIT = 400


class GenericCsvTransform:
    """Generic CSV File Type data for rows of data to be seperated by a delimeter."""

    filename: str
    rows: List[List[str]]
    delimiter: str = ","

    @staticmethod
    def build(
        filename: str, rows: List[List[str]], delimiter: str = ","
    ) -> "GenericCsvTransform":
        """Build a Generic CSV datatype class."""
        if "." in filename and not filename.endswith(".csv"):
            raise ValueError(
                f"Provided filename {filename} invalid. Only CSV file format is accepted."
            )
        elif "." not in filename:
            filename = f"{filename}.csv"
        csv = GenericCsvTransform()
        csv.filename = filename
        csv.rows = rows
        csv.delimiter = delimiter
        return csv


class ReadData(BaseModel):
    """Read Data type containing the wavelength for a Plate Reader read alongside the Measurement Data of that read."""

    wavelength: int
    data: Dict[str, float]


class PlateReaderData(BaseModel):
    """Data from a Opentrons Plate Reader Read. Can be converted to CSV template format."""

    read_results: List[ReadData]
    reference_wavelength: Optional[int] = None
    start_time: datetime
    finish_time: datetime
    serial_number: str

    def build_generic_csv(  # noqa: C901
        self, filename: str, measurement: ReadData
    ) -> GenericCsvTransform:
        """Builds a CSV compatible object containing Plate Reader Measurements.

        This will also automatically reformat the provided filename to include the wavelength of those measurements.
        """
        plate_alpharows = ["A", "B", "C", "D", "E", "F", "G", "H"]
        rows = []

        rows.append(["", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"])
        for i in range(8):
            row = [plate_alpharows[i]]
            for j in range(12):
                row.append(str(measurement.data[f"{plate_alpharows[i]}{j+1}"]))
            rows.append(row)
        for i in range(3):
            rows.append([])
        rows.append(["", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"])
        for i in range(8):
            row = [plate_alpharows[i]]
            for j in range(12):
                row.append("")
            rows.append(row)
        for i in range(3):
            rows.append([])
        rows.append(
            [
                "",
                "ID",
                "Well",
                "Absorbance (OD)",
                "Mean Absorbance (OD)",
                "Absorbance %CV",
            ]
        )
        for i in range(3):
            rows.append([])
        rows.append(
            [
                "",
                "ID",
                "Well",
                "Absorbance (OD)",
                "Mean Absorbance (OD)",
                "Dilution Factor",
                "Absorbance %CV",
            ]
        )
        rows.append(["1", "Sample 1", "", "", "", "1", "", "", "", "", "", ""])
        for i in range(3):
            rows.append([])

        # end of file metadata
        rows.append(["Protocol"])
        rows.append(["Assay"])
        rows.append(["Sample Wavelength (nm)", str(measurement.wavelength)])
        if self.reference_wavelength is not None:
            rows.append(["Reference Wavelength (nm)", str(self.reference_wavelength)])
        rows.append(["Serial No.", self.serial_number])
        rows.append(
            ["Measurement started at", self.start_time.strftime("%m %d %H:%M:%S %Y")]
        )
        rows.append(
            ["Measurement finished at", self.finish_time.strftime("%m %d %H:%M:%S %Y")]
        )

        # Ensure the filename adheres to ruleset contains the wavelength for a given measurement
        if filename.endswith(".csv"):
            filename = filename[:-4]
        filename = filename + str(measurement.wavelength) + "nm.csv"

        return GenericCsvTransform.build(
            filename=filename,
            rows=rows,
            delimiter=",",
        )


class FileProvider:
    """Provider class to wrap file read write interactions to the data files directory in the engine."""

    def __init__(
        self,
        data_files_write_csv_callback: Optional[
            Callable[[GenericCsvTransform], Awaitable[str]]
        ] = None,
        data_files_filecount: Optional[Callable[[], Awaitable[int]]] = None,
    ) -> None:
        """Initialize the interface callbacks of the File Provider for data file handling within the Protocol Engine.

        Params:
            data_files_write_csv_callback: Callback to write a CSV file to the data files directory and add it to the database.
            data_files_filecount: Callback to check the amount of data files already present in the data files directory.
        """
        self._data_files_write_csv_callback = data_files_write_csv_callback
        self._data_files_filecount = data_files_filecount

    async def write_csv(self, write_data: GenericCsvTransform) -> str:
        """Writes the provided CSV object to a file in the Data Files directory. Returns the File ID of the file created."""
        if self._data_files_filecount is not None:
            file_count = await self._data_files_filecount()
            if file_count >= MAXIMUM_CSV_FILE_LIMIT:
                raise StorageLimitReachedError(
                    f"Not enough space to store file {write_data.filename}."
                )
            if self._data_files_write_csv_callback is not None:
                return await self._data_files_write_csv_callback(write_data)
        # If we are in an analysis or simulation state, return an empty file ID
        return ""



"""Translate events from a legacy ``ProtocolContext`` into Protocol Engine commands."""

from collections import defaultdict
from datetime import datetime
from typing import Dict, List, Optional, Tuple, Union

from opentrons.hardware_control.modules.types import (
    ModuleModel as HardwareModuleModel,
    TemperatureModuleModel,
    MagneticModuleModel,
    ThermocyclerModuleModel,
    HeaterShakerModuleModel,
)
from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons.types import MountType, DeckSlotName, Location
from opentrons.legacy_commands import types as legacy_command_types
from opentrons.protocol_api import InstrumentContext
from opentrons.protocol_api.core.legacy.deck import FIXED_TRASH_ID
from opentrons.protocol_api.core.legacy.load_info import (
    LoadInfo as LegacyLoadInfo,
    LabwareLoadInfo as LegacyLabwareLoadInfo,
    InstrumentLoadInfo as LegacyInstrumentLoadInfo,
    ModuleLoadInfo as LegacyModuleLoadInfo,
)
from opentrons.protocol_engine import (
    ProtocolEngineError,
    actions as pe_actions,
    commands as pe_commands,
    types as pe_types,
)
from opentrons.protocol_engine.error_recovery_policy import ErrorRecoveryType
from opentrons.protocol_engine.resources import (
    ModelUtils,
    ModuleDataProvider,
    pipette_data_provider,
)
from opentrons.protocol_engine.state.update_types import (
    StateUpdate,
)

from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.errors import ErrorCodes, EnumeratedError, PythonException


class LegacyCommandParams(pe_commands.CustomParams):
    """Custom command data payload for mapped legacy commands."""

    legacyCommandType: str
    legacyCommandText: str


class LegacyContextCommandError(ProtocolEngineError):
    """An error returned when a PAPIv2 ProtocolContext command fails."""

    def __init__(self, wrapping_exc: BaseException) -> None:
        if isinstance(wrapping_exc, EnumeratedError):
            super().__init__(
                wrapping_exc.code,
                wrapping_exc.message,
                wrapping_exc.detail,
                wrapping_exc.wrapping,
            )
        else:
            super().__init__(
                code=ErrorCodes.GENERAL_ERROR,
                message=str(wrapping_exc),
                wrapping=[PythonException(wrapping_exc)],
            )


_HARDWARE_TO_PE_MODULE: Dict[HardwareModuleModel, pe_types.ModuleModel] = {
    MagneticModuleModel.MAGNETIC_V1: pe_types.ModuleModel.MAGNETIC_MODULE_V1,
    MagneticModuleModel.MAGNETIC_V2: pe_types.ModuleModel.MAGNETIC_MODULE_V2,
    TemperatureModuleModel.TEMPERATURE_V1: pe_types.ModuleModel.TEMPERATURE_MODULE_V1,
    TemperatureModuleModel.TEMPERATURE_V2: pe_types.ModuleModel.TEMPERATURE_MODULE_V2,
    ThermocyclerModuleModel.THERMOCYCLER_V1: pe_types.ModuleModel.THERMOCYCLER_MODULE_V1,
    ThermocyclerModuleModel.THERMOCYCLER_V2: pe_types.ModuleModel.THERMOCYCLER_MODULE_V2,
    HeaterShakerModuleModel.HEATER_SHAKER_V1: pe_types.ModuleModel.HEATER_SHAKER_MODULE_V1,
}

_HIGHER_ORDER_COMMAND_TYPES = {
    legacy_command_types.MIX,
    legacy_command_types.CONSOLIDATE,
    legacy_command_types.DISTRIBUTE,
    legacy_command_types.TRANSFER,
    legacy_command_types.RETURN_TIP,
    legacy_command_types.AIR_GAP,
}


class LegacyCommandMapper:
    """Map broker commands to protocol engine commands.

    Each protocol should use its own instance of this class.
    """

    def __init__(
        self, module_data_provider: Optional[ModuleDataProvider] = None
    ) -> None:
        """Initialize the command mapper."""
        # commands keyed by broker message ID
        self._commands_by_broker_id: Dict[str, pe_commands.Command] = {}

        # running count of each legacy command type, to construct IDs
        self._command_count: Dict[str, int] = defaultdict(lambda: 0)

        # equipment IDs by physical location
        self._labware_id_by_slot: Dict[DeckSlotName, str] = {
            DeckSlotName.FIXED_TRASH: FIXED_TRASH_ID
        }
        self._labware_id_by_module_id: Dict[str, str] = {}
        self._pipette_id_by_mount: Dict[MountType, str] = {}
        self._module_id_by_slot: Dict[DeckSlotName, str] = {}

        # module definition state and provider depedency
        self._module_definition_by_model: Dict[
            pe_types.ModuleModel, pe_types.ModuleDefinition
        ] = {}
        self._module_data_provider = module_data_provider or ModuleDataProvider()

    def map_command(  # noqa: C901
        self,
        command: legacy_command_types.CommandMessage,
    ) -> List[pe_actions.Action]:
        """Map a legacy Broker command to a ProtocolEngine command.

        A "before" message from the Broker
        is mapped to a ``RUNNING`` ProtocolEngine command.

        An "after" message from the Broker
        is mapped to a ``SUCCEEDED`` ProtocolEngine command.
        It has the same ID as the original ``RUNNING`` command,
        so when you send it to the ProtocolEngine, it will update the original
        command's status in-place.
        """
        command_type = command["name"]

        if command_type in _HIGHER_ORDER_COMMAND_TYPES:
            return []

        command_error = command["error"]
        stage = command["$"]
        # TODO(mc, 2021-12-08): use message ID as command ID directly once
        # https://github.com/Opentrons/opentrons/issues/8986 is resolved
        broker_id = command["id"]
        now = ModelUtils.get_timestamp()

        results: List[pe_actions.Action] = []

        if stage == "before":
            count = self._command_count[command_type]
            command_id = f"{command_type}-{count}"
            command_create, running_command = self._build_initial_command(
                command, command_id, now
            )

            self._command_count[command_type] = count + 1
            self._commands_by_broker_id[broker_id] = running_command

            results.append(
                pe_actions.QueueCommandAction(
                    command_id=command_id,
                    created_at=running_command.createdAt,
                    request=command_create,
                    request_hash=None,
                )
            )
            assert running_command.startedAt is not None
            results.append(
                pe_actions.RunCommandAction(
                    running_command.id, started_at=running_command.startedAt
                )
            )

        elif stage == "after":
            running_command = self._commands_by_broker_id[broker_id]
            completed_command: pe_commands.Command
            if command_error is None:
                if isinstance(running_command, pe_commands.PickUpTip):
                    completed_command = running_command.copy(
                        update={
                            "result": pe_commands.PickUpTipResult.construct(
                                tipVolume=command["payload"]["location"].max_volume,  # type: ignore[typeddict-item]
                                tipLength=command["payload"]["instrument"].hw_pipette["tip_length"],  # type: ignore[typeddict-item]
                                position=pe_types.DeckPoint(x=0, y=0, z=0),
                            ),
                            "status": pe_commands.CommandStatus.SUCCEEDED,
                            "completedAt": now,
                            "notes": [],
                        }
                    )
                elif isinstance(running_command, pe_commands.DropTip):
                    completed_command = running_command.copy(
                        update={
                            "result": pe_commands.DropTipResult.construct(
                                position=pe_types.DeckPoint(x=0, y=0, z=0)
                            ),
                            "status": pe_commands.CommandStatus.SUCCEEDED,
                            "completedAt": now,
                            "notes": [],
                        }
                    )
                elif isinstance(running_command, pe_commands.Aspirate):
                    completed_command = running_command.copy(
                        update={
                            # Don't .construct() result, because we want to validate
                            # volume.
                            "result": pe_commands.AspirateResult(
                                volume=running_command.params.volume,
                                position=pe_types.DeckPoint(x=0, y=0, z=0),
                            ),
                            "status": pe_commands.CommandStatus.SUCCEEDED,
                            "completedAt": now,
                            "notes": [],
                        }
                    )
                elif isinstance(running_command, pe_commands.Dispense):
                    completed_command = running_command.copy(
                        update={
                            # Don't .construct() result, because we want to validate
                            # volume.
                            "result": pe_commands.DispenseResult(
                                volume=running_command.params.volume,
                                position=pe_types.DeckPoint(x=0, y=0, z=0),
                            ),
                            "status": pe_commands.CommandStatus.SUCCEEDED,
                            "completedAt": now,
                            "notes": [],
                        }
                    )
                elif isinstance(running_command, pe_commands.BlowOut):
                    completed_command = running_command.copy(
                        update={
                            "result": pe_commands.BlowOutResult.construct(
                                position=pe_types.DeckPoint(x=0, y=0, z=0)
                            ),
                            "status": pe_commands.CommandStatus.SUCCEEDED,
                            "completedAt": now,
                            "notes": [],
                        }
                    )
                elif isinstance(running_command, pe_commands.Comment):
                    completed_command = running_command.copy(
                        update={
                            "result": pe_commands.CommentResult.construct(),
                            "status": pe_commands.CommandStatus.SUCCEEDED,
                            "completedAt": now,
                            "notes": [],
                        }
                    )
                elif isinstance(running_command, pe_commands.Custom):
                    completed_command = running_command.copy(
                        update={
                            "result": pe_commands.CustomResult.construct(),
                            "status": pe_commands.CommandStatus.SUCCEEDED,
                            "completedAt": now,
                            "notes": [],
                        }
                    )
                else:
                    # TODO(mm, 2024-06-13): This looks potentially wrong.
                    # We're creating a `SUCCEEDED` command that does not have a `result`,
                    # which is not normally possible.
                    completed_command = running_command.copy(
                        update={
                            "status": pe_commands.CommandStatus.SUCCEEDED,
                            "completedAt": now,
                            "notes": [],
                        }
                    )
                results.append(
                    pe_actions.SucceedCommandAction(
                        completed_command,
                        state_update=StateUpdate(),
                    )
                )

                if isinstance(completed_command, pe_commands.WaitForResume):
                    results.append(
                        pe_actions.PauseAction(source=pe_actions.PauseSource.PROTOCOL)
                    )

            else:
                results.append(
                    pe_actions.FailCommandAction(
                        command_id=running_command.id,
                        running_command=running_command,
                        error_id=ModelUtils.generate_id(),
                        failed_at=now,
                        error=LegacyContextCommandError(command_error),
                        notes=[],
                        # For legacy protocols, we don't attempt to support any kind
                        # of error recovery at the Protocol Engine level.
                        # These protocols only run on the OT-2, which doesn't have
                        # any recoverable errors, anyway.
                        type=ErrorRecoveryType.FAIL_RUN,
                    )
                )

        return results

    def map_equipment_load(self, load_info: LegacyLoadInfo) -> List[pe_actions.Action]:
        """Map a labware, instrument (pipette), or module load to a PE command."""
        if isinstance(load_info, LegacyLabwareLoadInfo):
            return self._map_labware_load(load_info)
        elif isinstance(load_info, LegacyInstrumentLoadInfo):
            return self._map_instrument_load(load_info)
        elif isinstance(load_info, LegacyModuleLoadInfo):
            return self._map_module_load(load_info)

    def _build_initial_command(
        self,
        command: legacy_command_types.CommandMessage,
        command_id: str,
        now: datetime,
    ) -> Tuple[pe_commands.CommandCreate, pe_commands.Command]:
        if command["name"] == legacy_command_types.PICK_UP_TIP:
            return self._build_pick_up_tip(
                command=command, command_id=command_id, now=now
            )
        elif command["name"] == legacy_command_types.DROP_TIP:
            return self._build_drop_tip(command=command, command_id=command_id, now=now)

        elif (
            command["name"] == legacy_command_types.ASPIRATE
            or command["name"] == legacy_command_types.DISPENSE
        ):
            return self._build_liquid_handling(
                command=command, command_id=command_id, now=now
            )
        elif command["name"] == legacy_command_types.BLOW_OUT:
            return self._build_blow_out(command=command, command_id=command_id, now=now)
        elif command["name"] == legacy_command_types.PAUSE:
            wait_for_resume_running = pe_commands.WaitForResume.construct(
                id=command_id,
                key=command_id,
                status=pe_commands.CommandStatus.RUNNING,
                createdAt=now,
                startedAt=now,
                params=pe_commands.WaitForResumeParams.construct(
                    message=command["payload"]["userMessage"],
                ),
            )
            wait_for_resume_create: pe_commands.CommandCreate = (
                pe_commands.WaitForResumeCreate.construct(
                    key=wait_for_resume_running.key,
                    params=wait_for_resume_running.params,
                )
            )
            return wait_for_resume_create, wait_for_resume_running
        elif command["name"] == legacy_command_types.COMMENT:
            comment_running = pe_commands.Comment.construct(
                id=command_id,
                key=command_id,
                status=pe_commands.CommandStatus.RUNNING,
                createdAt=now,
                startedAt=now,
                params=pe_commands.CommentParams.construct(
                    message=command["payload"]["text"],
                ),
            )
            comment_create = pe_commands.CommentCreate.construct(
                key=comment_running.key, params=comment_running.params
            )
            return comment_create, comment_running
        else:
            custom_running = pe_commands.Custom.construct(
                id=command_id,
                key=command_id,
                status=pe_commands.CommandStatus.RUNNING,
                createdAt=now,
                startedAt=now,
                params=LegacyCommandParams.construct(
                    legacyCommandType=command["name"],
                    legacyCommandText=command["payload"]["text"],
                ),
            )
            custom_create = pe_commands.CustomCreate.construct(
                key=custom_running.key,
                params=custom_running.params,
            )
            return custom_create, custom_running

    def _build_drop_tip(
        self,
        command: legacy_command_types.DropTipMessage,
        command_id: str,
        now: datetime,
    ) -> Tuple[pe_commands.CommandCreate, pe_commands.Command]:
        pipette: InstrumentContext = command["payload"]["instrument"]
        well = command["payload"]["location"]
        mount = MountType(pipette.mount)
        #   the following type checking suppression assumes the tiprack is not loaded on top of a module
        slot = DeckSlotName.from_primitive(well.parent.parent)  # type: ignore[arg-type]
        well_name = well.well_name
        labware_id = self._labware_id_by_slot[slot]
        pipette_id = self._pipette_id_by_mount[mount]

        running = pe_commands.DropTip.construct(
            id=command_id,
            key=command_id,
            status=pe_commands.CommandStatus.RUNNING,
            createdAt=now,
            startedAt=now,
            params=pe_commands.DropTipParams.construct(
                pipetteId=pipette_id,
                labwareId=labware_id,
                wellName=well_name,
            ),
        )
        create = pe_commands.DropTipCreate.construct(
            key=running.key,
            params=running.params,
        )
        return create, running

    def _build_pick_up_tip(
        self,
        command: legacy_command_types.PickUpTipMessage,
        command_id: str,
        now: datetime,
    ) -> Tuple[pe_commands.CommandCreate, pe_commands.Command]:
        pipette: InstrumentContext = command["payload"]["instrument"]
        location = command["payload"]["location"]
        well = location
        mount = MountType(pipette.mount)
        #   the following type checking suppression assumes the tiprack is not loaded on top of a module
        slot = DeckSlotName.from_primitive(well.parent.parent)  # type: ignore[arg-type]
        well_name = well.well_name
        labware_id = self._labware_id_by_slot[slot]
        pipette_id = self._pipette_id_by_mount[mount]

        running = pe_commands.PickUpTip.construct(
            id=command_id,
            key=command_id,
            status=pe_commands.CommandStatus.RUNNING,
            createdAt=now,
            startedAt=now,
            params=pe_commands.PickUpTipParams.construct(
                pipetteId=pipette_id,
                labwareId=labware_id,
                wellName=well_name,
            ),
        )
        create = pe_commands.PickUpTipCreate.construct(
            key=running.key, params=running.params
        )
        return create, running

    def _build_liquid_handling(
        self,
        command: Union[
            legacy_command_types.AspirateMessage, legacy_command_types.DispenseMessage
        ],
        command_id: str,
        now: datetime,
    ) -> Tuple[pe_commands.CommandCreate, pe_commands.Command]:
        pipette: InstrumentContext = command["payload"]["instrument"]
        location = command["payload"]["location"]
        volume = command["payload"]["volume"]
        # TODO:(jr, 15.08.2022): aspirate and dispense commands with no specified labware
        # get filtered into custom. Refactor this in followup legacy command mapping
        if location.labware.is_well:
            well = location.labware.as_well()
            slot = DeckSlotName(location.labware.first_parent())
            parent_module_id = self._module_id_by_slot.get(slot)
            labware_id = (
                self._labware_id_by_module_id[parent_module_id]
                if parent_module_id is not None
                else self._labware_id_by_slot[slot]
            )
            mount = MountType(pipette.mount)
            well_name = well.well_name
            pipette_id = self._pipette_id_by_mount[mount]

            if volume == 0:
                # In edge cases, it's possible for a Python protocol to do dispense()
                # or aspirate() with a volume of 0, which behaves roughly like
                # move_to(). Protocol Engine aspirate and dispense commands must have
                # volume > 0, so we can't map into those.
                #
                # TODO(mm, 2024-03-22): I don't think this has been true since
                # https://github.com/Opentrons/opentrons/pull/14211. Can we just use
                # aspirate and dispense commands now?
                move_to_well_running = pe_commands.MoveToWell.construct(
                    id=command_id,
                    key=command_id,
                    status=pe_commands.CommandStatus.RUNNING,
                    createdAt=now,
                    startedAt=now,
                    params=pe_commands.MoveToWellParams.construct(
                        pipetteId=pipette_id,
                        labwareId=labware_id,
                        wellName=well_name,
                    ),
                )
                move_to_well_create = pe_commands.MoveToWellCreate.construct(
                    key=move_to_well_running.key, params=move_to_well_running.params
                )
                return move_to_well_create, move_to_well_running
            elif command["name"] == legacy_command_types.ASPIRATE:
                flow_rate = command["payload"]["rate"] * pipette.flow_rate.aspirate
                aspirate_running = pe_commands.Aspirate.construct(
                    id=command_id,
                    key=command_id,
                    status=pe_commands.CommandStatus.RUNNING,
                    createdAt=now,
                    startedAt=now,
                    # Don't .construct() params, because we want to validate
                    # volume and flowRate.
                    params=pe_commands.AspirateParams(
                        pipetteId=pipette_id,
                        labwareId=labware_id,
                        wellName=well_name,
                        volume=volume,
                        flowRate=flow_rate,
                    ),
                )
                aspirate_create = pe_commands.AspirateCreate.construct(
                    key=aspirate_running.key, params=aspirate_running.params
                )
                return aspirate_create, aspirate_running
            else:
                flow_rate = command["payload"]["rate"] * pipette.flow_rate.dispense
                dispense_running = pe_commands.Dispense.construct(
                    id=command_id,
                    key=command_id,
                    status=pe_commands.CommandStatus.RUNNING,
                    createdAt=now,
                    startedAt=now,
                    # Don't .construct params, because we want to validate
                    # volume and flowRate.
                    params=pe_commands.DispenseParams(
                        pipetteId=pipette_id,
                        labwareId=labware_id,
                        wellName=well_name,
                        volume=volume,
                        flowRate=flow_rate,
                    ),
                )
                dispense_create = pe_commands.DispenseCreate.construct(
                    key=dispense_running.key, params=dispense_running.params
                )
                return dispense_create, dispense_running

        else:
            running = pe_commands.Custom.construct(
                id=command_id,
                key=command_id,
                status=pe_commands.CommandStatus.RUNNING,
                createdAt=now,
                startedAt=now,
                params=LegacyCommandParams.construct(
                    legacyCommandType=command["name"],
                    legacyCommandText=command["payload"]["text"],
                ),
            )
            create = pe_commands.CustomCreate.construct(
                key=running.key, params=running.params
            )
            return create, running

    def _build_blow_out(
        self,
        command: legacy_command_types.BlowOutMessage,
        command_id: str,
        now: datetime,
    ) -> Tuple[pe_commands.CommandCreate, pe_commands.Command]:
        pipette: InstrumentContext = command["payload"]["instrument"]
        location = command["payload"]["location"]
        flow_rate = pipette.flow_rate.blow_out
        #   TODO:(jr, 15.08.2022): blow_out commands with no specified labware get filtered
        #   into custom. Remove location.labware.is_empty is False when refactor is complete
        if isinstance(location, Location) and location.labware.is_well:
            well = location.labware.as_well()
            slot = DeckSlotName(location.labware.first_parent())
            parent_module_id = self._module_id_by_slot.get(slot)
            labware_id = (
                self._labware_id_by_module_id[parent_module_id]
                if parent_module_id is not None
                else self._labware_id_by_slot[slot]
            )
            mount = MountType(pipette.mount)
            well_name = well.well_name
            pipette_id = self._pipette_id_by_mount[mount]

            blow_out_running = pe_commands.BlowOut.construct(
                id=command_id,
                key=command_id,
                status=pe_commands.CommandStatus.RUNNING,
                createdAt=now,
                startedAt=now,
                # Don't .construct() params, because we want to validate flowRate.
                params=pe_commands.BlowOutParams(
                    pipetteId=pipette_id,
                    labwareId=labware_id,
                    wellName=well_name,
                    flowRate=flow_rate,
                ),
            )
            blow_out_create = pe_commands.BlowOutCreate.construct(
                key=blow_out_running.key, params=blow_out_running.params
            )
            return blow_out_create, blow_out_running

        #   TODO:(jr, 15.08.2022): blow_out commands with no specified labware get filtered
        #   into custom. Refactor this in followup legacy command mapping
        else:
            custom_running = pe_commands.Custom.construct(
                id=command_id,
                key=command_id,
                status=pe_commands.CommandStatus.RUNNING,
                createdAt=now,
                startedAt=now,
                params=LegacyCommandParams.construct(
                    legacyCommandType=command["name"],
                    legacyCommandText=command["payload"]["text"],
                ),
            )
            custom_create = pe_commands.CustomCreate.construct(
                key=custom_running.key, params=custom_running.params
            )
            return custom_create, custom_running

    def _map_labware_load(
        self, labware_load_info: LegacyLabwareLoadInfo
    ) -> List[pe_actions.Action]:
        """Map a legacy labware load to a ProtocolEngine command."""
        now = ModelUtils.get_timestamp()
        count = self._command_count["LOAD_LABWARE"]
        slot = labware_load_info.deck_slot
        location: pe_types.LabwareLocation
        if labware_load_info.on_module:
            location = pe_types.ModuleLocation.construct(
                moduleId=self._module_id_by_slot[slot]
            )
        else:
            location = pe_types.DeckSlotLocation.construct(slotName=slot)

        command_id = f"commands.LOAD_LABWARE-{count}"
        labware_id = f"labware-{count}"

        succeeded_command = pe_commands.LoadLabware.construct(
            id=command_id,
            key=command_id,
            status=pe_commands.CommandStatus.SUCCEEDED,
            createdAt=now,
            startedAt=now,
            completedAt=now,
            params=pe_commands.LoadLabwareParams.construct(
                location=location,
                loadName=labware_load_info.labware_load_name,
                namespace=labware_load_info.labware_namespace,
                version=labware_load_info.labware_version,
                displayName=labware_load_info.labware_display_name,
            ),
            notes=[],
            result=pe_commands.LoadLabwareResult.construct(
                labwareId=labware_id,
                definition=LabwareDefinition.parse_obj(
                    labware_load_info.labware_definition
                ),
                offsetId=labware_load_info.offset_id,
            ),
        )
        queue_action = pe_actions.QueueCommandAction(
            command_id=succeeded_command.id,
            created_at=succeeded_command.createdAt,
            request=pe_commands.LoadLabwareCreate.construct(
                key=succeeded_command.key, params=succeeded_command.params
            ),
            request_hash=None,
        )
        run_action = pe_actions.RunCommandAction(
            command_id=succeeded_command.id,
            # We just set this above, so we know it's not None.
            started_at=succeeded_command.startedAt,  # type: ignore[arg-type]
        )
        state_update = StateUpdate()
        assert succeeded_command.result is not None
        state_update.set_loaded_labware(
            labware_id=labware_id,
            definition=succeeded_command.result.definition,
            display_name=labware_load_info.labware_display_name,
            offset_id=labware_load_info.offset_id,
            location=location,
        )

        succeed_action = pe_actions.SucceedCommandAction(
            command=succeeded_command,
            state_update=state_update,
        )

        self._command_count["LOAD_LABWARE"] = count + 1
        if isinstance(location, pe_types.DeckSlotLocation):
            self._labware_id_by_slot[location.slotName] = labware_id
        elif isinstance(location, pe_types.ModuleLocation):
            self._labware_id_by_module_id[location.moduleId] = labware_id

        return [queue_action, run_action, succeed_action]

    def _map_instrument_load(
        self,
        instrument_load_info: LegacyInstrumentLoadInfo,
    ) -> List[pe_actions.Action]:
        """Map a legacy instrument (pipette) load to a ProtocolEngine command.

        Also creates a `AddPipetteConfigAction`, which is not necessary for the run,
        but is needed for stop so tip geometry is in state for the HardwareStopper.
        """
        now = ModelUtils.get_timestamp()
        count = self._command_count["LOAD_PIPETTE"]
        command_id = f"commands.LOAD_PIPETTE-{count}"
        pipette_id = f"pipette-{count}"
        mount = MountType(str(instrument_load_info.mount).lower())

        succeeded_command = pe_commands.LoadPipette.construct(
            id=command_id,
            key=command_id,
            status=pe_commands.CommandStatus.SUCCEEDED,
            createdAt=now,
            startedAt=now,
            completedAt=now,
            params=pe_commands.LoadPipetteParams.construct(
                pipetteName=PipetteNameType(instrument_load_info.instrument_load_name),
                mount=mount,
            ),
            notes=[],
            result=pe_commands.LoadPipetteResult.construct(pipetteId=pipette_id),
        )
        serial = instrument_load_info.pipette_dict.get("pipette_id", None) or ""
        state_update = StateUpdate()
        state_update.set_load_pipette(
            pipette_id=pipette_id,
            mount=succeeded_command.params.mount,
            pipette_name=succeeded_command.params.pipetteName,
            liquid_presence_detection=succeeded_command.params.liquidPresenceDetection,
        )
        state_update.update_pipette_config(
            pipette_id=pipette_id,
            serial_number=serial,
            config=pipette_data_provider.get_pipette_static_config(
                # Compatibility note - this is the version of tip overlap data, it stays at 0
                # so protocol behavior does not change when you run a legacy JSON protocol
                instrument_load_info.pipette_dict,
                "v0",
            ),
        )
        queue_action = pe_actions.QueueCommandAction(
            command_id=succeeded_command.id,
            created_at=succeeded_command.createdAt,
            request=pe_commands.LoadPipetteCreate.construct(
                key=succeeded_command.key, params=succeeded_command.params
            ),
            request_hash=None,
        )
        run_action = pe_actions.RunCommandAction(
            command_id=succeeded_command.id,
            # We just set this above, so we know it's not None.
            started_at=succeeded_command.startedAt,  # type: ignore[arg-type]
        )

        succeed_action = pe_actions.SucceedCommandAction(
            command=succeeded_command,
            state_update=state_update,
        )

        self._command_count["LOAD_PIPETTE"] = count + 1
        self._pipette_id_by_mount[mount] = pipette_id

        return [queue_action, run_action, succeed_action]

    def _map_module_load(
        self, module_load_info: LegacyModuleLoadInfo
    ) -> List[pe_actions.Action]:
        """Map a legacy module load to a Protocol Engine command."""
        now = ModelUtils.get_timestamp()

        count = self._command_count["LOAD_MODULE"]
        command_id = f"commands.LOAD_MODULE-{count}"
        module_id = f"module-{count}"
        requested_model = _HARDWARE_TO_PE_MODULE[module_load_info.requested_model]
        loaded_model = _HARDWARE_TO_PE_MODULE[module_load_info.loaded_model]

        # This will fetch a V2 definition only. PAPI < v2.3 use V1 definitions.
        # When running a < v2.3 protocol, there will be a mismatch of definitions used
        # during analysis+LPC (V2) and protocol execution (V1).
        # But this shouldn't result in any problems since V2 and V1 definitions
        # have similar info, with V2 having additional info fields.
        loaded_definition = self._module_definition_by_model.get(
            loaded_model
        ) or self._module_data_provider.get_definition(loaded_model)

        succeeded_command = pe_commands.LoadModule.construct(
            id=command_id,
            key=command_id,
            status=pe_commands.CommandStatus.SUCCEEDED,
            createdAt=now,
            startedAt=now,
            completedAt=now,
            params=pe_commands.LoadModuleParams.construct(
                model=requested_model,
                location=pe_types.DeckSlotLocation(
                    slotName=module_load_info.deck_slot,
                ),
                moduleId=module_id,
            ),
            notes=[],
            result=pe_commands.LoadModuleResult.construct(
                moduleId=module_id,
                serialNumber=module_load_info.module_serial,
                definition=loaded_definition,
                model=loaded_model,
            ),
        )
        queue_action = pe_actions.QueueCommandAction(
            command_id=succeeded_command.id,
            created_at=succeeded_command.createdAt,
            request=pe_commands.LoadModuleCreate.construct(
                key=succeeded_command.key, params=succeeded_command.params
            ),
            request_hash=None,
        )
        run_action = pe_actions.RunCommandAction(
            command_id=succeeded_command.id,
            # We just set this above, so we know it's not None.
            started_at=succeeded_command.startedAt,  # type: ignore[arg-type]
        )
        succeed_action = pe_actions.SucceedCommandAction(
            command=succeeded_command, state_update=StateUpdate()
        )

        self._command_count["LOAD_MODULE"] = count + 1
        self._module_id_by_slot[module_load_info.deck_slot] = module_id
        self._module_definition_by_model[loaded_model] = loaded_definition

        return [queue_action, run_action, succeed_action]



from typing import List

from .types import PauseType


class PauseManager:
    """This class determines whether or not the hardware controller should
    pause or resume by evaluating the pause and resume types. The use of two
    pause types are used to separate the delay resume (triggered when the delay
    timer runs out) and the pause resume (trigged by user via the app).
    """

    def __init__(self) -> None:
        self.queue: List[PauseType] = []

    @property
    def should_pause(self) -> bool:
        return bool(self.queue)

    def resume(self, pause_type: PauseType) -> None:
        try:
            self.queue.remove(pause_type)
        except ValueError:
            pass

    def pause(self, pause_type: PauseType) -> None:
        if pause_type not in self.queue:
            self.queue.append(pause_type)

    def reset(self) -> None:
        self.queue = []



from opentrons.drivers.types import Temperature
from opentrons.hardware_control.modules.types import TemperatureStatus


class LidTemperatureStatus:
    """A wrapper for the lid temperature status."""

    TEMP_THRESHOLD = 0.3
    """The threshold under which the difference between target and temperature
    is considered `holding`."""

    def __init__(self) -> None:
        """Construct."""
        self._status = TemperatureStatus.ERROR

    @property
    def status(self) -> TemperatureStatus:
        """Return the current status."""
        return self._status

    def update(self, temperature: Temperature) -> TemperatureStatus:
        """Update the status based on the temperature."""
        if temperature.target is None:
            status = TemperatureStatus.IDLE
        else:
            diff = temperature.target - temperature.current
            if abs(diff) < self.TEMP_THRESHOLD:
                status = TemperatureStatus.HOLDING
            elif diff < 0:
                # TC lid can't actively cool
                status = TemperatureStatus.IDLE
            else:
                status = TemperatureStatus.HEATING
        self._status = status
        return self._status



"""Heater-Shaker Module sub-state."""
from dataclasses import dataclass
from typing import NewType, Optional

from opentrons.protocol_engine.types import (
    TemperatureRange,
    SpeedRange,
    HeaterShakerLatchStatus,
)
from opentrons.protocol_engine.errors import (
    InvalidTargetTemperatureError,
    InvalidTargetSpeedError,
    NoTargetTemperatureSetError,
    CannotPerformModuleAction,
)

HeaterShakerModuleId = NewType("HeaterShakerModuleId", str)


# TODO (spp, 2022-03-22): Move these values to heater-shaker module definition.
HEATER_SHAKER_TEMPERATURE_RANGE = TemperatureRange(min=37, max=95)
HEATER_SHAKER_SPEED_RANGE = SpeedRange(min=200, max=3000)


@dataclass(frozen=True)
class HeaterShakerModuleSubState:
    """Heater-Shaker-specific state.

    Provides calculations and read-only state access
    for an individual loaded Heater-Shaker Module.
    """

    module_id: HeaterShakerModuleId
    labware_latch_status: HeaterShakerLatchStatus
    is_plate_shaking: bool
    plate_target_temperature: Optional[float]

    def get_plate_target_temperature(self) -> float:
        """Get the module's target plate temperature."""
        target = self.plate_target_temperature

        if target is None:
            raise NoTargetTemperatureSetError(
                f"Module {self.module_id} does not have a target temperature set."
            )
        return target

    @staticmethod
    def validate_target_temperature(celsius: float) -> float:
        """Verify that the target temperature being set is valid for heater-shaker."""
        if (
            HEATER_SHAKER_TEMPERATURE_RANGE.min
            <= celsius
            <= HEATER_SHAKER_TEMPERATURE_RANGE.max
        ):
            return celsius
        else:
            raise InvalidTargetTemperatureError(
                f"Cannot set Heater-Shaker to {celsius} °C."
                f" Valid range is {HEATER_SHAKER_TEMPERATURE_RANGE} °C."
            )

    @staticmethod
    def validate_target_speed(rpm: float) -> int:
        """Verify that the target speed is valid for heater-shaker & convert to int."""
        rpm_int = int(round(rpm, 0))
        if HEATER_SHAKER_SPEED_RANGE.min <= rpm <= HEATER_SHAKER_SPEED_RANGE.max:
            return rpm_int
        else:
            raise InvalidTargetSpeedError(
                f"Cannot set Heater-Shaker to shake at {rpm} rpm. Valid speed range is "
                f"{HEATER_SHAKER_SPEED_RANGE} rpm."
            )

    def raise_if_labware_latch_not_closed(self) -> None:
        """Raise an error if labware is not latched on the heater-shaker."""
        if self.labware_latch_status == HeaterShakerLatchStatus.UNKNOWN:
            raise CannotPerformModuleAction(
                "Heater-Shaker cannot start or deactivate shaking if the labware latch has not been set to closed."
            )
        elif self.labware_latch_status == HeaterShakerLatchStatus.OPEN:
            raise CannotPerformModuleAction(
                "Heater-Shaker cannot start or deactivate shaking while the labware latch is open."
            )

    def raise_if_shaking(self) -> None:
        """Raise an error if the heater-shaker is currently shaking."""
        if self.is_plate_shaking:
            raise CannotPerformModuleAction(
                "Heater-Shaker cannot open its labware latch while it is shaking."
            )



import abc
import asyncio
import logging
import re
from typing import ClassVar, Mapping, Optional, TypeVar, cast
from packaging.version import InvalidVersion, parse, Version
from opentrons.config import IS_ROBOT, ROBOT_FIRMWARE_DIR
from opentrons.drivers.rpi_drivers.types import USBPort

from ..execution_manager import ExecutionManager
from .types import (
    BundledFirmware,
    ModuleDisconnectedCallback,
    UploadFunction,
    LiveData,
    ModuleType,
)

mod_log = logging.getLogger(__name__)

TaskPayload = TypeVar("TaskPayload")


def parse_fw_version(version: str) -> Version:
    try:
        device_version = parse(version)
        # This is a patch for older versions of packaging - they would try and parse old
        # kidns of versions and return a LegacyVersion object. We can't check for that
        # explicitly because they removed it in modern versions of packaging.
        if not isinstance(device_version, Version):
            raise InvalidVersion()
    except InvalidVersion:
        device_version = parse("v0.0.0")
    return cast(Version, device_version)


class AbstractModule(abc.ABC):
    """Defines the common methods of a module."""

    MODULE_TYPE: ClassVar[ModuleType]

    @classmethod
    @abc.abstractmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        poll_interval_seconds: Optional[float] = None,
        simulating: bool = False,
        sim_model: Optional[str] = None,
        sim_serial_number: Optional[str] = None,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ) -> "AbstractModule":
        """Modules should always be created using this factory.

        This lets the (perhaps blocking) work of connecting to and initializing
        a module be in a place that can be async.
        """

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ) -> None:
        self._port = port
        self._usb_port = usb_port
        self._loop = hw_control_loop
        self._execution_manager = execution_manager
        self._bundled_fw: Optional[BundledFirmware] = self.get_bundled_fw()
        self._disconnected_callback = disconnected_callback
        self._updating = False

    @staticmethod
    def sort_key(inst: "AbstractModule") -> int:
        usb_port = inst.usb_port

        primary_port = usb_port.port_number

        if usb_port.hub_port is not None:
            secondary_port = usb_port.hub_port
        else:
            secondary_port = 0

        return primary_port * 1000 + secondary_port

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def updating(self) -> bool:
        """The device is updating is True."""
        return self._updating

    def disconnected_callback(self) -> None:
        """Called from within the module object to signify the object is no longer connected"""
        if self._disconnected_callback is not None:
            self._disconnected_callback(self.port, self.serial_number)

    def get_bundled_fw(self) -> Optional[BundledFirmware]:
        """Get absolute path to bundled version of module fw if available."""
        if not IS_ROBOT:
            return None
        file_prefix = self.firmware_prefix()

        MODULE_FW_RE = re.compile(f"^{file_prefix}@v(.*)[.](hex|bin|byoup)$")
        for fw_resource in ROBOT_FIRMWARE_DIR.iterdir():  # type: ignore
            matches = MODULE_FW_RE.search(fw_resource.name)
            if matches:
                return BundledFirmware(version=matches.group(1), path=fw_resource)

        mod_log.info(f"no available fw file found for: {file_prefix}")
        return None

    def has_available_update(self) -> bool:
        """Return whether a newer firmware file is available"""
        if self.device_info and self._bundled_fw:
            device_version = parse_fw_version(self.device_info["version"])
            available_version = parse_fw_version(self._bundled_fw.version)
            return available_version > device_version
        return False

    async def wait_for_is_running(self) -> None:
        if not self.is_simulated:
            await self._execution_manager.wait_for_is_running()

    def make_cancellable(self, task: "asyncio.Task[TaskPayload]") -> None:
        self._execution_manager.register_cancellable_task(task)

    @abc.abstractmethod
    async def deactivate(self, must_be_running: bool = True) -> None:
        """Deactivate the module.

        Contains an override to the `wait_for_is_running` step in cases where the
        module must be deactivated regardless of context."""
        pass

    @property
    @abc.abstractmethod
    def status(self) -> str:
        """Return some string describing status."""
        pass

    @property
    @abc.abstractmethod
    def device_info(self) -> Mapping[str, str]:
        """Return a dict of the module's static information (serial, etc)"""
        pass

    @property
    @abc.abstractmethod
    def live_data(self) -> LiveData:
        """Return a dict of the module's dynamic information"""
        pass

    @property
    @abc.abstractmethod
    def is_simulated(self) -> bool:
        """True if >this is a simulated module."""
        pass

    @property
    def port(self) -> str:
        """The virtual port where the module is connected."""
        return self._port

    @property
    def usb_port(self) -> USBPort:
        """The physical port where the module is connected."""
        return self._usb_port

    @property
    def serial_number(self) -> Optional[str]:
        """The usb serial number of this device."""
        return self.device_info.get("serial")

    @abc.abstractmethod
    async def prep_for_update(self) -> str:
        """Prepare for an update.

        By the time this coroutine completes, the hardware should be ready
        to take an update. This implicitly tears down the module instance;
        it does not need to be either working or recoverable after this
        coroutine completes.

        :returns str: The port we're running on.
        """
        pass

    @property
    def bundled_fw(self) -> Optional[BundledFirmware]:
        return self._bundled_fw

    @abc.abstractmethod
    def model(self) -> str:
        """A name for this specific module, matching module defs"""
        pass

    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        """A shortname used for matching usb ports, among other things"""
        pass

    @abc.abstractmethod
    def firmware_prefix(self) -> str:
        """The prefix used for looking up firmware"""
        pass

    @abc.abstractmethod
    def bootloader(self) -> UploadFunction:
        """Method used to upload file to this module's bootloader."""
        pass

    async def cleanup(self) -> None:
        """Clean up the module instance.

        Clean up, i.e. stop pollers, disconnect serial, etc in preparation for
        object destruction.
        """
        pass



"""Protocol engine exceptions."""

from __future__ import annotations

from logging import getLogger
from typing import Any, Dict, Final, Optional, Union, Iterator, Sequence, TYPE_CHECKING

from opentrons_shared_data.errors import ErrorCodes
from opentrons_shared_data.errors.exceptions import EnumeratedError, PythonException

if TYPE_CHECKING:
    from opentrons.protocol_engine.types import TipGeometry


log = getLogger(__name__)


class ProtocolEngineError(EnumeratedError):
    """Base Protocol Engine error class."""

    def __init__(
        self,
        code: Optional[ErrorCodes] = None,
        message: Optional[str] = None,
        detail: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ProtocolEngineError."""
        super().__init__(
            code=code or ErrorCodes.GENERAL_ERROR,
            message=message,
            detail=detail,
            wrapping=wrapping,
        )


class UnexpectedProtocolError(ProtocolEngineError):
    """Raised when an unexpected error occurs.

    This error is indicative of a software bug. If it happens, it means an
    exception was raised somewhere in the stack and it was not properly caught
    and wrapped.
    """

    def __init__(
        self,
        message: Optional[str] = None,
        wrapping: Optional[Sequence[Union[EnumeratedError, BaseException]]] = None,
    ) -> None:
        """Initialize an UnexpectedProtocolError with an original error."""

        def _convert_exc() -> Iterator[EnumeratedError]:
            if not wrapping:
                return
            for exc in wrapping:
                if isinstance(exc, EnumeratedError):
                    yield exc
                else:
                    yield PythonException(exc)

        super().__init__(
            code=ErrorCodes.GENERAL_ERROR,
            message=message,
            wrapping=[e for e in _convert_exc()],
        )


# TODO(mc, 2020-10-18): differentiate between pipette missing vs incorrect.
# By comparison, loadModule uses ModuleAlreadyPresentError and ModuleNotAttachedError.
class FailedToLoadPipetteError(ProtocolEngineError):
    """Raised when a LoadPipette command fails.

    This failure may be caused by:
    - An incorrect pipette already attached to the mount
    - A missing pipette on the requested mount
    """

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FailedToLoadPipetteError."""
        super().__init__(ErrorCodes.PIPETTE_NOT_PRESENT, message, details, wrapping)


# TODO(mc, 2020-10-18): differentiate between pipette missing vs incorrect
class PipetteNotAttachedError(ProtocolEngineError):
    """Raised when an operation's required pipette is not attached."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a PipetteNotAttachedError."""
        super().__init__(ErrorCodes.PIPETTE_NOT_PRESENT, message, details, wrapping)


class InvalidLoadPipetteSpecsError(ProtocolEngineError):
    """Raised when a loadPipette uses invalid specifications."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an InvalidLoadPipetteSpecsError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidSpecificationForRobotTypeError(ProtocolEngineError):
    """Raised when a command provides invalid specs for the given robot type."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an InvalidSpecificationForRobotTypeError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class TipNotAttachedError(ProtocolEngineError):
    """Raised when an operation's required pipette tip is not attached."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a PipetteNotAttachedError."""
        super().__init__(ErrorCodes.UNEXPECTED_TIP_REMOVAL, message, details, wrapping)


class PickUpTipTipNotAttachedError(TipNotAttachedError):
    """Raised from TipHandler.pick_up_tip().

    This is like TipNotAttachedError except that it carries some extra information
    about the attempted operation.
    """

    tip_geometry: Final[TipGeometry]
    """The tip geometry that would have been on the pipette, had the operation succeeded."""

    def __init__(self, tip_geometry: TipGeometry) -> None:
        super().__init__()
        self.tip_geometry = tip_geometry


class TipAttachedError(ProtocolEngineError):
    """Raised when a tip shouldn't be attached, but is."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a PIpetteNotAttachedError."""
        super().__init__(ErrorCodes.UNEXPECTED_TIP_ATTACH, message, details, wrapping)


class CommandDoesNotExistError(ProtocolEngineError):
    """Raised when referencing a command that does not exist."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CommandDoesNotExistError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareNotLoadedError(ProtocolEngineError):
    """Raised when referencing a labware that has not been loaded."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareNotLoadedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareNotLoadedOnModuleError(ProtocolEngineError):
    """Raised when there is no labware loaded on the requested module."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareNotLoadedOnModuleError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareNotLoadedOnLabwareError(ProtocolEngineError):
    """Raised when there is no labware loaded on the requested labware."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareNotLoadedOnLabwareError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareNotOnDeckError(ProtocolEngineError):
    """Raised when a labware can't be used because it's off-deck."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareNotOnDeckError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LiquidDoesNotExistError(ProtocolEngineError):
    """Raised when referencing a liquid that has not been loaded."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LiquidDoesNotExistError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareDefinitionDoesNotExistError(ProtocolEngineError):
    """Raised when referencing a labware definition that does not exist."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareDefinitionDoesNotExistError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareCannotBeStackedError(ProtocolEngineError):
    """Raised when trying to load labware onto another labware it is not defined to be loaded onto."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareCannotBeStackedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareIsInStackError(ProtocolEngineError):
    """Raised when trying to move to or physically interact with a labware that has another labware on top."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareIsInStackError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareOffsetDoesNotExistError(ProtocolEngineError):
    """Raised when referencing a labware offset that does not exist."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareOffsetDoesNotExistError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareIsNotTipRackError(ProtocolEngineError):
    """Raised when trying to use a regular labware as a tip rack."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareIsNotTiprackError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareIsTipRackError(ProtocolEngineError):
    """Raised when trying to use a command not allowed on tip rack."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareIsTiprackError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareIsAdapterError(ProtocolEngineError):
    """Raised when trying to use a command not allowed on adapter."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareIsAdapterError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class TouchTipDisabledError(ProtocolEngineError):
    """Raised when touch tip is used on well with touchTipDisabled quirk."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a TouchTipDisabledError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class WellDoesNotExistError(ProtocolEngineError):
    """Raised when referencing a well that does not exist."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a WellDoesNotExistError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class PipetteNotLoadedError(ProtocolEngineError):
    """Raised when referencing a pipette that has not been loaded."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a PipetteNotLoadedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class ModuleNotLoadedError(ProtocolEngineError):
    """Raised when referencing a module that has not been loaded."""

    def __init__(self, *, module_id: str) -> None:
        super().__init__(ErrorCodes.GENERAL_ERROR, f"Module {module_id} not found.")


class ModuleNotOnDeckError(ProtocolEngineError):
    """Raised when trying to use a module that is loaded off the deck."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ModuleNotOnDeckError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class ModuleNotConnectedError(ProtocolEngineError):
    """Raised when trying to use a module that is not connected to the robot electrically."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ModuleNotConnectedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class SlotDoesNotExistError(ProtocolEngineError):
    """Raised when referencing a deck slot that does not exist."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a SlotDoesNotExistError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class CutoutDoesNotExistError(ProtocolEngineError):
    """Raised when referencing a cutout that does not exist."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CutoutDoesNotExistError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class FixtureDoesNotExistError(ProtocolEngineError):
    """Raised when referencing a cutout fixture that does not exist."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FixtureDoesNotExist."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class AddressableAreaDoesNotExistError(ProtocolEngineError):
    """Raised when referencing an addressable area that does not exist."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a AddressableAreaDoesNotExistError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class FixtureDoesNotProvideAreasError(ProtocolEngineError):
    """Raised when a cutout fixture does not provide any addressable areas for a requested cutout."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FixtureDoesNotProvideAreasError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class AreaNotInDeckConfigurationError(ProtocolEngineError):
    """Raised when an addressable area is referenced that is not provided by a deck configuration."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a AreaNotInDeckConfigurationError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class IncompatibleAddressableAreaError(ProtocolEngineError):
    """Raised when two non-compatible addressable areas are referenced during analysis."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a IncompatibleAddressableAreaError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


# TODO(mc, 2020-11-06): flesh out with structured data to replicate
# existing LabwareHeightError
class FailedToPlanMoveError(ProtocolEngineError):
    """Raised when a requested movement could not be planned."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FailedToPlanmoveError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class MustHomeError(ProtocolEngineError):
    """Raised when motors must be homed due to unknown current position."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a MustHomeError."""
        super().__init__(ErrorCodes.POSITION_UNKNOWN, message, details, wrapping)


class CommandNotAllowedError(ProtocolEngineError):
    """Raised when adding a command with bad data."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CommandNotAllowedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class FixitCommandNotAllowedError(ProtocolEngineError):
    """Raised when adding a fixit command to a non-recoverable engine."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a SetupCommandNotAllowedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class SetupCommandNotAllowedError(ProtocolEngineError):
    """Raised when adding a setup command to a non-idle/non-paused engine."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a SetupCommandNotAllowedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class ResumeFromRecoveryNotAllowedError(ProtocolEngineError):
    """Raised when attempting to resume a run from recovery that has a fixit command in the queue."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ResumeFromRecoveryNotAllowedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class PauseNotAllowedError(ProtocolEngineError):
    """Raised when attempting to pause a run that is not running."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a PauseNotAllowedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class RunStoppedError(ProtocolEngineError):
    """Raised when attempting to interact with a stopped engine."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a RunStoppedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class ModuleNotAttachedError(ProtocolEngineError):
    """Raised when a requested module is not attached."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ModuleNotAttached."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class ModuleAlreadyPresentError(ProtocolEngineError):
    """Raised when a module is already present in a requested location."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ModuleAlreadyPresentError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class WrongModuleTypeError(ProtocolEngineError):
    """Raised when performing a module action on the wrong kind of module."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a WrongModuleTypeError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class ThermocyclerNotOpenError(ProtocolEngineError):
    """Raised when trying to move to a labware that's in a closed Thermocycler."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ThermocyclerNotOpenError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class RobotDoorOpenError(ProtocolEngineError):
    """Raised when executing a protocol command when a robot door is open."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a RobotDoorOpenError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class PipetteMovementRestrictedByHeaterShakerError(ProtocolEngineError):
    """Raised when trying to move to labware that's restricted by a module."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a PipetteMovementRestrictedByHeaterShakerError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class HeaterShakerLabwareLatchNotOpenError(ProtocolEngineError):
    """Raised when Heater-Shaker latch is not open when it is expected to be so."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a HeaterShakerLabwareLatchNotOpenError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class HeaterShakerLabwareLatchStatusUnknown(ProtocolEngineError):
    """Raised when Heater-Shaker latch has not been set before moving to it."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a HeaterShakerLabwareLatchStatusUnknown."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class EngageHeightOutOfRangeError(ProtocolEngineError):
    """Raised when a Magnetic Module engage height is out of bounds."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a EngageHeightOutOfRangeError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class NoMagnetEngageHeightError(ProtocolEngineError):
    """Raised if a Magnetic Module engage height is missing."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a NoMagnetEngageHeightError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class NoTargetTemperatureSetError(ProtocolEngineError):
    """Raised when awaiting temperature when no target was set."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a NoTargetTemperatureSetError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidTargetTemperatureError(ProtocolEngineError):
    """Raised when attempting to set an invalid target temperature."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a InvalidTargetTemperatureError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidBlockVolumeError(ProtocolEngineError):
    """Raised when attempting to set an invalid block max volume."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a InvalidBlockVolumeError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidWavelengthError(ProtocolEngineError):
    """Raised when attempting to set an invalid absorbance wavelength."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a InvalidWavelengthError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidHoldTimeError(ProtocolEngineError):
    """An error raised when attempting to set an invalid temperature hold time."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a InvalidHoldTimeError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidTargetSpeedError(ProtocolEngineError):
    """Raised when attempting to set an invalid target speed."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a InvalidTargetSpeedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class CannotPerformModuleAction(ProtocolEngineError):
    """Raised when trying to perform an illegal hardware module action."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CannotPerformModuleAction."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class HardwareNotSupportedError(ProtocolEngineError):
    """Raised when executing a command on the wrong hardware."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a HardwareNotSupportedError."""
        super().__init__(
            ErrorCodes.NOT_SUPPORTED_ON_ROBOT_TYPE, message, details, wrapping
        )


class GripperNotAttachedError(ProtocolEngineError):
    """Raised when executing a gripper action without an attached gripper."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a GripperNotAttachedError."""
        super().__init__(ErrorCodes.GRIPPER_NOT_PRESENT, message, details, wrapping)


class CannotPerformGripperAction(ProtocolEngineError):
    """Raised when trying to perform an illegal gripper action."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CannotPerformGripperAction."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareMovementNotAllowedError(ProtocolEngineError):
    """Raised when attempting an illegal labware movement."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareMovementNotAllowedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LabwareIsNotAllowedInLocationError(ProtocolEngineError):
    """Raised when attempting an illegal labware load into slot."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareIsNotAllowedInLocationError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LocationIsOccupiedError(ProtocolEngineError):
    """Raised when attempting to place labware in a non-empty location."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LocationIsOccupiedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LocationNotAccessibleByPipetteError(ProtocolEngineError):
    """Raised when attempting to move pipette to an inaccessible location."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LocationNotAccessibleByPipetteError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LocationIsStagingSlotError(ProtocolEngineError):
    """Raised when referencing a labware on a staging slot when trying to get standard deck slot."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LocationIsStagingSlotError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LocationIsLidDockSlotError(ProtocolEngineError):
    """Raised when referencing a labware on a lid dock slot when trying to get standard deck slot."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LocationIsLidDockSlotError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class FirmwareUpdateRequired(ProtocolEngineError):
    """Raised when the firmware needs to be updated."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LocationIsOccupiedError."""
        super().__init__(
            ErrorCodes.FIRMWARE_UPDATE_REQUIRED, message, details, wrapping
        )


class PipetteNotReadyToAspirateError(ProtocolEngineError):
    """Raised when the pipette is not ready to aspirate."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a PipetteNotReadyToAspirateError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidAspirateVolumeError(ProtocolEngineError):
    """Raised when pipetting a volume larger than the pipette volume."""

    def __init__(
        self,
        attempted_aspirate_volume: float,
        available_volume: float,
        max_pipette_volume: float,
        max_tip_volume: Optional[float],  # None if there's no tip.
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a InvalidPipettingVolumeError."""
        message = (
            f"Cannot aspirate {attempted_aspirate_volume} µL when only"
            f" {available_volume} is available in the tip."
        )
        details = {
            "attempted_aspirate_volume": attempted_aspirate_volume,
            "available_volume": available_volume,
            "max_pipette_volume": max_pipette_volume,
            "max_tip_volume": max_tip_volume,
        }
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidDispenseVolumeError(ProtocolEngineError):
    """Raised when attempting to dispense a volume that was not aspirated."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a InvalidDispenseVolumeError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidPushOutVolumeError(ProtocolEngineError):
    """Raised when attempting to use an invalid volume for dispense push_out."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a InvalidPushOutVolumeError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidAxisForRobotType(ProtocolEngineError):
    """Raised when attempting to use an axis that is not present on the given type of robot."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a InvalidAxisForRobotType."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class InvalidLiquidHeightFound(ProtocolEngineError):
    """Raised when attempting to estimate liquid height based on volume fails."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an InvalidLiquidHeightFound error."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class LiquidHeightUnknownError(ProtocolEngineError):
    """Raised when attempting to specify WellOrigin.MENISCUS before liquid probing has been done."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LiquidHeightUnknownError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class EStopActivatedError(ProtocolEngineError):
    """Represents an E-stop event."""

    def __init__(
        self,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an EStopActivatedError."""
        super().__init__(
            code=ErrorCodes.E_STOP_ACTIVATED,
            message="E-stop activated.",
            wrapping=wrapping,
        )


class NotSupportedOnRobotType(ProtocolEngineError):
    """Raised when attempting to perform an action that is not supported for the given robot type."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a NotSupportedOnRobotType exception."""
        super().__init__(
            ErrorCodes.NOT_SUPPORTED_ON_ROBOT_TYPE, message, details, wrapping
        )


class TipNotEmptyError(ProtocolEngineError):
    """Raised when an operation requires an empty tip but is provided a tip with liquid."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a TipNotEmptyError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class IncompleteLabwareDefinitionError(ProtocolEngineError):
    """Raised when a labware definition lacks innerLabwareGeometry in general or for a specific well_id."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an IncompleteLabwareDefinitionError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class IncompleteWellDefinitionError(ProtocolEngineError):
    """Raised when a well definition lacks a geometryDefinitionId."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an IncompleteWellDefinitionError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class OperationLocationNotInWellError(ProtocolEngineError):
    """Raised when a calculated operation location is not within a well."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an OperationLocationNotInWellError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)


class StorageLimitReachedError(ProtocolEngineError):
    """Raised to indicate that a file cannot be created due to storage limitations."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an StorageLimitReached."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, detail, wrapping)



"""Parameter context for python protocols."""
import uuid
from typing import List, Optional, Union, Dict

from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import requires_version
from opentrons.protocols.parameters import (
    parameter_definition,
    csv_parameter_definition,
    validation,
)
from opentrons.protocols.parameters.types import (
    ParameterChoice,
    UserFacingTypes,
)
from opentrons.protocols.parameters.exceptions import (
    ParameterDefinitionError,
    ParameterValueError,
    IncompatibleParameterError,
)
from opentrons.protocol_engine.types import (
    RunTimeParameter,
    PrimitiveRunTimeParamValuesType,
    CSVRuntimeParamPaths,
    FileInfo,
)

from ._parameters import Parameters

_ParameterDefinitionTypes = Union[
    parameter_definition.ParameterDefinition,
    csv_parameter_definition.CSVParameterDefinition,
]


class ParameterContext:
    """Public context for adding parameters to a protocol."""

    def __init__(self, api_version: APIVersion) -> None:
        """Initializes a parameter context for user-set parameters."""
        self._api_version = api_version
        self._parameters: Dict[str, _ParameterDefinitionTypes] = {}

    def add_int(
        self,
        display_name: str,
        variable_name: str,
        default: int,
        minimum: Optional[int] = None,
        maximum: Optional[int] = None,
        choices: Optional[List[ParameterChoice]] = None,
        description: Optional[str] = None,
        unit: Optional[str] = None,
    ) -> None:
        """Creates an integer parameter, settable within a given range or list of choices.

        Arguments:
            display_name: The display name of the int parameter as it would show up on the frontend.
            variable_name: The variable name the int parameter will be referred to in the run context.
            default: The default value the int parameter will be set to. This will be used in initial analysis.
            minimum: The minimum value the int parameter can be set to (inclusive). Mutually exclusive with choices.
            maximum: The maximum value the int parameter can be set to (inclusive). Mutually exclusive with choices.
            choices: A list of possible choices that this parameter can be set to.
                Mutually exclusive with minimum and maximum.
            description: A description of the parameter as it will show up on the frontend.
            unit: An optional unit to be appended to the end of the integer as it shown on the frontend.
        """
        validation.validate_variable_name_unique(variable_name, set(self._parameters))
        parameter = parameter_definition.create_int_parameter(
            display_name=display_name,
            variable_name=variable_name,
            default=default,
            minimum=minimum,
            maximum=maximum,
            choices=choices,
            description=description,
            unit=unit,
        )
        self._parameters[parameter.variable_name] = parameter

    def add_float(
        self,
        display_name: str,
        variable_name: str,
        default: float,
        minimum: Optional[float] = None,
        maximum: Optional[float] = None,
        choices: Optional[List[ParameterChoice]] = None,
        description: Optional[str] = None,
        unit: Optional[str] = None,
    ) -> None:
        """Creates a float parameter, settable within a given range or list of choices.

        Arguments:
            display_name: The display name of the float parameter as it would show up on the frontend.
            variable_name: The variable name the float parameter will be referred to in the run context.
            default: The default value the float parameter will be set to. This will be used in initial analysis.
            minimum: The minimum value the float parameter can be set to (inclusive). Mutually exclusive with choices.
            maximum: The maximum value the float parameter can be set to (inclusive). Mutually exclusive with choices.
            choices: A list of possible choices that this parameter can be set to.
                Mutually exclusive with minimum and maximum.
            description: A description of the parameter as it will show up on the frontend.
            unit: An optional unit to be appended to the end of the float as it shown on the frontend.
        """
        validation.validate_variable_name_unique(variable_name, set(self._parameters))
        parameter = parameter_definition.create_float_parameter(
            display_name=display_name,
            variable_name=variable_name,
            default=validation.ensure_float_value(default),
            minimum=validation.ensure_optional_float_value(minimum),
            maximum=validation.ensure_optional_float_value(maximum),
            choices=validation.ensure_float_choices(choices),
            description=description,
            unit=unit,
        )
        self._parameters[parameter.variable_name] = parameter

    def add_bool(
        self,
        display_name: str,
        variable_name: str,
        default: bool,
        description: Optional[str] = None,
    ) -> None:
        """Creates a boolean parameter with allowable values of "On" (True) or "Off" (False).

        Arguments:
            display_name: The display name of the boolean parameter as it would show up on the frontend.
            variable_name: The variable name the boolean parameter will be referred to in the run context.
            default: The default value the boolean parameter will be set to. This will be used in initial analysis.
            description: A description of the parameter as it will show up on the frontend.
        """
        validation.validate_variable_name_unique(variable_name, set(self._parameters))
        parameter = parameter_definition.create_bool_parameter(
            display_name=display_name,
            variable_name=variable_name,
            default=default,
            choices=[
                {"display_name": "On", "value": True},
                {"display_name": "Off", "value": False},
            ],
            description=description,
        )
        self._parameters[parameter.variable_name] = parameter

    def add_str(
        self,
        display_name: str,
        variable_name: str,
        default: str,
        choices: Optional[List[ParameterChoice]] = None,
        description: Optional[str] = None,
    ) -> None:
        """Creates a string parameter, settable among given choices.

        Arguments:
            display_name: The display name of the string parameter as it would show up on the frontend.
            variable_name: The variable name the string parameter will be referred to in the run context.
            default: The default value the string parameter will be set to. This will be used in initial analysis.
            choices: A list of possible choices that this parameter can be set to.
                Mutually exclusive with minimum and maximum.
            description: A description of the parameter as it will show up on the frontend.
        """
        validation.validate_variable_name_unique(variable_name, set(self._parameters))
        parameter = parameter_definition.create_str_parameter(
            display_name=display_name,
            variable_name=variable_name,
            default=default,
            choices=choices,
            description=description,
        )
        self._parameters[parameter.variable_name] = parameter

    @requires_version(2, 20)
    def add_csv_file(
        self,
        display_name: str,
        variable_name: str,
        description: Optional[str] = None,
    ) -> None:
        """Creates a CSV parameter that can be set to any CSV file via the Flex ODD or a local computer.

        Arguments:
            display_name: The display name of the CSV parameter as it would show up on the frontend.
            variable_name: The variable name the CSV parameter will be referred to in the run context.
            description: A description of the parameter as it will show up on the frontend.
        """
        if any(
            isinstance(parameter, csv_parameter_definition.CSVParameterDefinition)
            for parameter in self._parameters.values()
        ):
            raise IncompatibleParameterError(
                "Only one CSV File parameter can be defined per protocol."
            )

        validation.validate_variable_name_unique(variable_name, set(self._parameters))
        parameter = csv_parameter_definition.create_csv_parameter(
            display_name=display_name,
            variable_name=variable_name,
            description=description,
        )
        self._parameters[parameter.variable_name] = parameter

    def set_parameters(
        self, parameter_overrides: PrimitiveRunTimeParamValuesType
    ) -> None:
        """Sets parameters to values given by client, validating them as well.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        for variable_name, override_value in parameter_overrides.items():
            try:
                parameter = self._parameters[variable_name]
            except KeyError:
                raise ParameterDefinitionError(
                    f"Parameter {variable_name} is not defined as a parameter for this protocol."
                )
            if isinstance(parameter, csv_parameter_definition.CSVParameterDefinition):
                raise ParameterValueError(
                    f"A primitive param value was provided for the parameter '{variable_name}',"
                    f" but '{variable_name}' is a CSV parameter that can only accept file IDs."
                )
            else:
                validated_value = validation.ensure_value_type(
                    override_value, parameter.parameter_type
                )
                parameter.value = validated_value

    def initialize_csv_files(
        self, run_time_param_file_overrides: CSVRuntimeParamPaths
    ) -> None:
        """Initializes the files for CSV parameters.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        for variable_name, file_path in run_time_param_file_overrides.items():
            try:
                parameter = self._parameters[variable_name]
            except KeyError:
                raise ParameterDefinitionError(
                    f"Parameter {variable_name} is not defined as a parameter for this protocol."
                )
            if not isinstance(
                parameter, csv_parameter_definition.CSVParameterDefinition
            ):
                raise ParameterValueError(
                    f"File Id was provided for the parameter '{variable_name}',"
                    f" but '{variable_name}' is not a CSV parameter."
                )

            # TODO(jbl 2024-09-30) Refactor this so file ID is passed as its own argument and not assumed from the path
            # If this is running on a robot, the parent folder in the path will be the file ID
            # If it is running locally, most likely the parent folder will not be a UUID, so instead we will change
            # this to be an empty string
            file_id = file_path.parent.name
            try:
                uuid.UUID(file_id, version=4)
            except ValueError:
                file_id = ""

            file_name = file_path.name

            with file_path.open("rb") as fh:
                contents = fh.read()

            parameter.file_info = FileInfo(id=file_id, name=file_name)
            parameter.value = contents

    def export_parameters_for_analysis(self) -> List[RunTimeParameter]:
        """Exports all parameters into a protocol engine models for reporting in analysis.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return [
            parameter.as_protocol_engine_type()
            for parameter in self._parameters.values()
        ]

    def export_parameters_for_protocol(self) -> Parameters:
        """Exports all parameters into a protocol run usable parameters object.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        parameters_for_protocol: Dict[str, UserFacingTypes] = {}
        for parameter in self._parameters.values():
            value: UserFacingTypes
            if isinstance(parameter, csv_parameter_definition.CSVParameterDefinition):
                value = parameter.as_csv_parameter_interface(self._api_version)
            else:
                value = parameter.value
            parameters_for_protocol[parameter.variable_name] = value
        return Parameters(parameters=parameters_for_protocol)
